from rest_framework import fields, serializers
from rest_framework.exceptions import ValidationError

from django.conf import settings
from django.db import IntegrityError
from django.db.models import Count

from haupt.apis.serializers.artifacts import (
    RunArtifactBackwardCompatibleSerializer,
    RunArtifactSerializer,
)
from haupt.apis.serializers.base.owner_mixin import ProjectOwnerMixin
from haupt.apis.serializers.base.project import ProjectMixin
from haupt.apis.serializers.base.tags import TagsMixin
from haupt.apis.serializers.base.uuid_slug_related_field import UuidSlugRelatedField
from haupt.db.defs import Models
from haupt.db.managers.versions import get_component_version_state
from polyaxon._config.spec import ConfigSpec
from polyaxon._constants.metadata import META_IS_PROMOTED
from polyaxon._polyaxonfile import ComponentSpecification
from polyaxon.schemas import V1ProjectVersionKind


class ProjectVersionStageSerializer(serializers.ModelSerializer):
    uuid = fields.UUIDField(format="hex", read_only=True)
    condition = fields.DictField(
        write_only=True, allow_empty=True, allow_null=True, required=False
    )

    class Meta:
        model = Models.ProjectVersion
        fields = ("uuid", "stage", "condition", "stage_conditions")
        extra_kwargs = {
            "stage": {"read_only": True},
            "stage_conditions": {"read_only": True},
        }


class ProjectVersionNameSerializer(serializers.ModelSerializer):
    class Meta:
        model = Models.ProjectVersion
        fields = ("name",)


class ProjectVersionSerializer(serializers.ModelSerializer):
    uuid = fields.UUIDField(format="hex", read_only=True)
    state = fields.UUIDField(format="hex", read_only=True)

    class Meta:
        model = Models.ProjectVersion
        fields = (
            "uuid",
            "name",
            "description",
            "tags",
            "created_at",
            "updated_at",
            "stage",
            "state",
            "kind",
        )

    def _validate_run(self, run, owner):
        org_filters = {"project__owner": owner} if settings.HAS_ORG_MANAGEMENT else {}
        if run and Models.Run.objects.filter(**org_filters, uuid=run).count() != 1:
            raise ValidationError(
                f"The run({run}) does not exist,"
                f"not valid,"
                f"or not part of this organization."
            )


class ProjectVersionDetailSerializer(
    ProjectVersionSerializer,
    ProjectMixin,
    ProjectOwnerMixin,
    TagsMixin,
):
    meta_info = fields.SerializerMethodField()
    owner = fields.SerializerMethodField()
    project = fields.SerializerMethodField()
    run = UuidSlugRelatedField(
        slug_field="uuid",
        queryset=Models.Run.objects.select_related("project").only(
            "id", "name", "project__id", "project__name"
        ),
        required=False,
        allow_null=True,
    )
    artifacts = serializers.SerializerMethodField()

    _LINEAGE_SERIALIZER = RunArtifactSerializer

    class Meta(ProjectVersionSerializer.Meta):
        fields = ProjectVersionSerializer.Meta.fields + (
            "owner",
            "project",
            "content",
            "readme",
            "run",
            "artifacts",
            "meta_info",
        )

    def get_meta_info(self, obj: Models.ProjectVersion):
        info = {}
        if obj.run_id:
            info["run"] = {
                "project": obj.run.project.name,
                "name": obj.run.name,
                "uuid": obj.run.uuid.hex,
            }
        if settings.HAS_ORG_MANAGEMENT and obj.connection_id:
            info["connection"] = {
                "uuid": obj.connection.uuid.hex,
                "name": "{}/{}".format(obj.connection.agent.name, obj.connection.name),
            }

        lineage = obj.lineage.all()
        compatibility = self.context.get("compatibility")

        if compatibility in ["old", "both"]:
            artifacts = [
                RunArtifactBackwardCompatibleSerializer(a).data for a in lineage
            ]
            if artifacts:
                info["artifacts"] = artifacts
        if not compatibility or compatibility in ["new", "both"]:
            lineage = [
                self._LINEAGE_SERIALIZER(a, context={"run": obj.run}).data
                for a in lineage
            ]
            if lineage:
                info["lineage"] = lineage
        return info

    def get_artifacts(self, obj: Models.ProjectVersion):
        return list(obj.lineage.values_list("artifact__name", flat=True))

    def validated_content(self, kind, content):
        if kind == V1ProjectVersionKind.COMPONENT:
            if not content:
                raise serializers.ValidationError(
                    "the `content` is required to create a component version."
                )
            try:
                return ComponentSpecification.read(content)
            except Exception as e:
                raise serializers.ValidationError(e)
        else:
            if content:
                try:
                    return ConfigSpec.read_from(content, config_type=".yaml")
                except Exception as e:
                    raise serializers.ValidationError(e)

    def validated_project_run(self, project, run):
        if not run or run.project_id == project.id:
            return

        if settings.HAS_ORG_MANAGEMENT and run.project.owner_id != project.owner_id:
            raise ValidationError(
                "The run referenced is not accessible from your organization."
            )

        projects = set(project.projects.values_list("id", flat=True))
        if projects:
            projects.add(project.id)
        if projects and run.project_id not in projects:
            raise ValidationError(
                "The run referenced is not accessible from this project."
            )

    def validated_project_connection(self, project, connection):
        if not connection:
            return

        if connection.owner_id != project.owner_id:
            raise ValidationError(
                "The connection referenced is not accessible from your organization."
            )

        connections = set(project.connections.values_list("id", flat=True))
        if connections and connection.id not in connections:
            raise ValidationError(
                "The connection referenced is not accessible from this project."
            )

    def validated_project_artifacts(self, run, artifacts):
        if artifacts is None:
            return None
        if not artifacts:
            return artifacts
        if not run:
            raise ValidationError(
                "A run is required to link the artifacts: `{}`.".format(artifacts)
            )
        artifacts = set(artifacts)
        found = run.artifacts_lineage.filter(artifact__name__in=artifacts).values_list(
            "id", "artifact_id", "artifact__name"
        )
        missing = artifacts - set(f[2] for f in found)
        if missing:
            raise ValidationError(
                "Some artifacts could not be found: `{}`.".format(missing)
            )
        return set(f[0] for f in found)

    def update(self, instance, validated_data):
        kind = validated_data.pop("kind", None)
        if kind and instance.kind != kind:
            raise ValidationError("Changing a version kind is not permitted.")
        content = validated_data.get("content")
        if content:
            json_content = self.validated_content(instance.kind, content)
            if kind == V1ProjectVersionKind.COMPONENT:
                validated_data["state"] = get_component_version_state(
                    component=json_content
                )
        current_run_id = instance.run_id
        validated_data = self.validated_tags(
            validated_data=validated_data, tags=instance.tags
        )
        self.validated_project_run(instance.project, validated_data.get("run"))
        if settings.HAS_ORG_MANAGEMENT:
            self.validated_project_connection(
                instance.project, validated_data.get("connection")
            )
        lineage = self.validated_project_artifacts(
            validated_data.get("run", instance.run if instance.run_id else None),
            validated_data.pop("artifacts", None),
        )

        try:
            instance = super().update(instance=instance, validated_data=validated_data)
        except IntegrityError:
            raise ValidationError(
                f"A version with name {validated_data['name']} already exists "
                f"on the requested project."
            )

        if lineage is not None:
            instance.lineage.set(lineage)

        new_run_id = instance.run_id
        if current_run_id != new_run_id:
            runs = Models.Run.restorable.filter(id__in=[current_run_id, new_run_id])
            runs = runs.annotate(versions_count=Count("versions"))
            runs = runs.only("id", "meta_info")
            runs_by_id = {run.id: run for run in runs}
            if (
                current_run_id
                and current_run_id in runs_by_id
                and runs_by_id[current_run_id].versions_count == 0
            ):
                runs_by_id[current_run_id].meta_info.pop(META_IS_PROMOTED, None)
            if (
                new_run_id
                and new_run_id in runs_by_id
                and not runs_by_id[new_run_id].meta_info.get(META_IS_PROMOTED)
            ):
                runs_by_id[new_run_id].meta_info[META_IS_PROMOTED] = True
            Models.Run.objects.bulk_update(runs, ["meta_info"])

        return instance

    def create(self, validated_data):
        kind = None
        if hasattr(self, "get_kind"):
            kind = self.get_kind()
        if not kind:
            kind = validated_data.pop("kind", None)
        if not kind:
            raise ValidationError("A project version kind is required.")
        validated_data["kind"] = kind
        content = validated_data.get("content")
        json_content = self.validated_content(kind, content)
        if kind == V1ProjectVersionKind.COMPONENT:
            validated_data["state"] = get_component_version_state(
                component=json_content
            )

        project = validated_data["project"]
        self.validated_project_run(project, validated_data.get("run"))
        if settings.HAS_ORG_MANAGEMENT:
            self.validated_project_connection(project, validated_data.get("connection"))
        lineage = self.validated_project_artifacts(
            validated_data.get("run"), validated_data.pop("artifacts", None)
        )
        try:
            instance = super().create(validated_data)
        except IntegrityError:
            raise ValidationError(
                f"A version with name {validated_data['name']} already exists "
                f"on the requested project."
            )

        if lineage is not None:
            instance.lineage.set(lineage)

        if instance.run_id:
            run = Models.Run.restorable.only("id", "meta_info").get(id=instance.run_id)
            if not run.meta_info.get(META_IS_PROMOTED):
                run.meta_info[META_IS_PROMOTED] = True
                run.save(update_fields=["meta_info"])

        return instance


class ProjectWriteVersionDetailSerializer(ProjectVersionDetailSerializer):
    kind = fields.SerializerMethodField()
    artifacts = serializers.ListField(
        child=serializers.CharField(allow_null=True, required=False),
        allow_null=True,
        required=False,
        write_only=True,
    )


class ComponentVersionDetailSerializer(ProjectWriteVersionDetailSerializer):
    def get_kind(self, obj=None):
        return V1ProjectVersionKind.COMPONENT


class ModelVersionDetailSerializer(ProjectWriteVersionDetailSerializer):
    def get_kind(self, obj=None):
        return V1ProjectVersionKind.MODEL


class ArtifactVersionDetailSerializer(ProjectWriteVersionDetailSerializer):
    def get_kind(self, obj=None):
        return V1ProjectVersionKind.ARTIFACT



import uuid

from collections import namedtuple
from typing import Optional


class OptionOwners(namedtuple("OptionOwners", "user project team organization")):
    @classmethod
    def get_owners(
        cls,
        user: Optional[int] = None,
        project: Optional[int] = None,
        team: Optional[int] = None,
        organization: Optional[int] = None,
    ) -> "OptionOwners":
        return cls(user=user, project=project, team=team, organization=organization)

    def to_dict(self):
        return dict(self._asdict())

    def __eq__(self, other):
        return (
            self.user == other.user
            and self.project == other.project
            and self.team == other.team
            and self.organization == other.organization
        )

    def __str__(self):
        return uuid.uuid5(
            namespace=uuid.NAMESPACE_DNS,
            name=f"user<{self.user}>:"
            f"project<{self.project}>:"
            f"team<{self.team}>:"
            f"organization<{self.organization}>",
        ).hex



from rest_framework import serializers

from haupt.db.defs import Models


class StatsUserSerializerMixin(serializers.ModelSerializer):
    LIGHT = True
    user = serializers.SerializerMethodField()

    class Meta:
        fields = (
            "created_at",
            "updated_at",
            "user",
            "run",
            "status",
            "version",
            "tracking_time",
        )

    def get_user(self, obj):
        if not obj.user:
            return obj.user
        if self.LIGHT:
            return {"count": obj.user.get("count", 0)}
        if not obj.user.get("ids"):
            return obj.user
        users = Models.User.objects.filter(id__in=obj.user["ids"]).values_list(
            "username", flat=True
        )
        return {"count": obj.user["count"], "users": list(users)}


class ProjectStatsSerializer(StatsUserSerializerMixin):
    class Meta(StatsUserSerializerMixin.Meta):
        model = Models.ProjectStats



from rest_framework.urlpatterns import format_suffix_patterns

from django.urls import re_path

from haupt.apis.run_lineage import views
from haupt.common.apis.urls import runs

run_lineages_urlpatterns = [
    re_path(
        runs.URLS_RUNS_ARTIFACTS_LINEAGE_LIST_V0, views.RunArtifactListView.as_view()
    ),
    re_path(
        runs.URLS_RUNS_ARTIFACTS_LINEAGE_NAMES_V0,
        views.RunArtifactNameListView.as_view(),
    ),
    re_path(
        runs.URLS_RUNS_ARTIFACTS_LINEAGE_DETAILS_V0,
        views.RunArtifactDetailView.as_view(),
    ),
    re_path(runs.URLS_RUNS_CLONES_LINEAGE_LIST, views.RunClonesListView.as_view()),
    re_path(runs.URLS_RUNS_UPSTREAM_LINEAGE_LIST, views.RunUpstreamListView.as_view()),
    re_path(
        runs.URLS_RUNS_DOWNSTREAM_LINEAGE_LIST, views.RunDownstreamListView.as_view()
    ),
    re_path(
        runs.URLS_RUNS_EDGES_LINEAGE_CREATE, views.SetRunEdgesLineageView.as_view()
    ),
    re_path(runs.URLS_RUNS_ARTIFACTS_LINEAGE_LIST, views.RunArtifactListView.as_view()),
    re_path(
        runs.URLS_RUNS_ARTIFACTS_LINEAGE_NAMES, views.RunArtifactNameListView.as_view()
    ),
    re_path(
        runs.URLS_RUNS_ARTIFACTS_LINEAGE_DETAILS, views.RunArtifactDetailView.as_view()
    ),
]

# Order is important, because the patterns could swallow other urls
urlpatterns = format_suffix_patterns(run_lineages_urlpatterns)



import uuid

from django.db import models


class UuidModel(models.Model):
    uuid = models.UUIDField(default=uuid.uuid4, editable=False, unique=True, null=False)

    class Meta:
        abstract = True



from rest_framework.urlpatterns import format_suffix_patterns

from django.urls import re_path

from haupt.apis.agents import views
from haupt.common.apis.urls import agents

options_urlpatterns = [
    re_path(agents.URLS_CATALOGS_AGENTS_STATE, views.AgentStateViewV1.as_view()),
    re_path(agents.URLS_CATALOGS_AGENTS_CRON, views.AgentCronViewV1.as_view()),
    re_path(agents.URLS_ORGANIZATIONS_RUNS_DETAILS, views.AgentRunDetailView.as_view()),
]

urlpatterns = format_suffix_patterns(options_urlpatterns)



from typing import Optional

from clipped.compact.pydantic import ValidationError
from clipped.formatting import Printer

from haupt.schemas.sandbox_config import SandboxConfig
from polyaxon._services.values import PolyaxonServices

PROXIES_CONFIG = None
SANDBOX_CONFIG = None

PolyaxonServices.set_service_name()


def set_proxies_config():
    from haupt.managers.proxies import ProxiesManager

    global PROXIES_CONFIG

    PROXIES_CONFIG = ProxiesManager.get_config_from_env()


def set_sandbox_config(
    config: Optional[SandboxConfig] = None,
    path: Optional[str] = None,
    persist: bool = False,
    env_only_config: bool = True,
):
    from haupt.managers.sandbox import SandboxConfigManager
    from polyaxon.settings import HOME_CONFIG, set_agent_config

    SandboxConfigManager.set_config_path(HOME_CONFIG.path)
    PolyaxonServices.set_service_name(PolyaxonServices.SANDBOX)

    global SANDBOX_CONFIG

    try:
        SANDBOX_CONFIG = (
            config or SandboxConfigManager.get_config_from_env()
            if env_only_config
            else SandboxConfigManager.get_config_or_default()
        )
        SANDBOX_CONFIG.mount_sandbox(path=path or HOME_CONFIG.path)
        SANDBOX_CONFIG.set_default_artifacts_store()
        if persist:
            SandboxConfigManager.set_config(SANDBOX_CONFIG)
        set_agent_config(SANDBOX_CONFIG)

    except (TypeError, ValidationError):
        SandboxConfigManager.purge()
        Printer.warning("Your sandbox configuration was purged!")



from haupt.db.administration.utils import DiffModelAdmin, ReadOnlyAdmin


class RunLightAdmin(DiffModelAdmin, ReadOnlyAdmin):
    list_display = (
        "uuid",
        "user",
        "project",
        "name",
        "status",
        "created_at",
        "updated_at",
        "started_at",
        "finished_at",
    )
    fields = (
        "project",
        "name",
        "description",
        "status",
        "live_state",
        "created_at",
        "updated_at",
        "started_at",
        "finished_at",
    )
    readonly_fields = ("status",)



from typing import List, Optional

import click


@click.command(
    context_settings=dict(
        ignore_unknown_options=True,
        allow_extra_args=True,
    )
)
@click.option(
    "--app",
    default="haupt",
)
@click.option(
    "-c",
    "--command",
    help="The command to execute.",
)
@click.option(
    "--path",
    help="The sandbox root path.",
)
@click.argument("args", nargs=-1)
def manage(app: Optional[str], command: str, path: str, args: Optional[List[str]]):
    """Start a new sever session."""
    if app and app != "haupt":
        from haupt.cli.runners.manage import manage

        return manage(command, args, app=app)

    from haupt.cli.runners.manage import run_manage
    from polyaxon._cli.config import set_home_path

    if path:
        set_home_path(home_path=path)

    return run_manage(command, args)



from hashlib import md5 as _md5
from hashlib import sha1 as _sha1

try:
    from django.utils.encoding import force_bytes
except ImportError:
    raise ImportError("This module depends on django.")


def md5_text(*args):
    m = _md5()
    for x in args:
        m.update(force_bytes(x, errors="replace"))
    return m


def sha1_text(*args):
    m = _sha1()
    for x in args:
        m.update(force_bytes(x, errors="replace"))
    return m



from django.db import models
from django.utils.timezone import now


class RunPipelines(models.Model):
    pipeline = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name="pipeline_runs",
    )
    controller = models.ForeignKey(
        "self",
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name="controller_runs",
    )
    upstream_runs = models.ManyToManyField(
        "self",
        blank=True,
        symmetrical=False,
        through="db.RunEdge",
        related_name="downstream_runs",
    )
    schedule_at = models.DateTimeField(blank=True, null=True, db_index=True)
    checked_at = models.DateTimeField(blank=True, null=True, default=now, db_index=True)

    class Meta:
        abstract = True



from typing import Optional, Tuple

from clipped.utils.enums import PEnum

from haupt.common.options.exceptions import OptionException
from polyaxon._config.parser import ConfigParser

NAMESPACE_DB_OPTION_MARKER = ":"
NAMESPACE_DB_CONFIG_MARKER = "__"
NAMESPACE_SETTINGS_MARKER = "__"
NAMESPACE_ENV_MARKER = ":"


class OptionStores(str, PEnum):
    ENV = "env"
    DB_OPTION = "db_option"
    DB_CONFIG = "db_config"
    SETTINGS = "settings"


class OptionScope(str, PEnum):
    GLOBAL = "global"
    ORGANISATION = "organization"
    TEAM = "team"
    PROJECT = "project"
    USER = "user"


class Option:
    key = None
    scope = OptionScope.GLOBAL
    is_secret = False
    is_optional = True
    is_list = False
    store = None
    typing = None
    default = None
    options = None
    description = None
    cache_ttl = 0

    @staticmethod
    def get_default_value():
        return None

    @classmethod
    def default_value(cls):
        return cls.default if cls.default is not None else cls.get_default_value()

    @classmethod
    def is_global(cls):
        return cls.scope == OptionScope.GLOBAL

    @classmethod
    def get_marker(cls) -> str:
        if cls.store == OptionStores.DB_OPTION:
            return NAMESPACE_DB_OPTION_MARKER
        if cls.store == OptionStores.DB_CONFIG:
            return NAMESPACE_DB_CONFIG_MARKER
        if cls.store == OptionStores.SETTINGS:
            return NAMESPACE_SETTINGS_MARKER

        return NAMESPACE_ENV_MARKER

    @classmethod
    def parse_key(cls) -> Tuple[Optional[str], str]:
        marker = cls.get_marker()
        parts = cls.key.split(marker)
        if len(parts) > 2:
            raise OptionException(
                "Option declared with multi-namespace key `{}`.".format(cls.key)
            )
        if len(parts) == 1:
            return None, cls.key
        return parts[0], parts[1]

    @classmethod
    def get_namespace(cls) -> Optional[str]:
        return cls.parse_key()[0]

    @classmethod
    def get_key_subject(cls):
        return cls.parse_key()[1]

    @classmethod
    def to_dict(cls, value=None):
        return {
            "key": cls.key,
            "typing": cls.typing,
            "is_list": cls.is_list,
            "is_secret": cls.is_secret,
            "value": value if value is not None else cls.default,
            "description": cls.description,
        }

    @classmethod
    def _extra_processing(cls, value):
        return value

    @classmethod
    def parse(cls, value):
        _value = ConfigParser.parse(cls.typing)(
            key=cls.key,
            value=value,
            is_list=cls.is_list,
            is_optional=cls.is_optional,
            default=cls.default,
            options=cls.options,
        )
        return cls._extra_processing(_value)



import logging

from typing import List, Optional

from haupt.schemas.platform_config import PlatformConfig


def set_ui(context, config: PlatformConfig, processors: Optional[List[str]] = None):
    logging.debug(config.config_module)
    context["ROOT_URLCONF"] = "{}.urls".format(config.config_module)
    platform_host = config.platform_host
    context["PLATFORM_HOST"] = platform_host

    def get_allowed_hosts():
        allowed_hosts = config.allowed_hosts
        if platform_host:
            allowed_hosts.append(platform_host)
        if ".polyaxon.com" not in allowed_hosts:
            allowed_hosts.append(".polyaxon.com")
        pod_ip = config.pod_ip
        if pod_ip:
            allowed_hosts.append(pod_ip)
        host_ip = config.host_ip
        if host_ip:
            host_cidr = ".".join(host_ip.split(".")[:-1])
            allowed_hosts += ["{}.{}".format(host_cidr, i) for i in range(255)]

        return allowed_hosts

    context["ALLOWED_HOSTS"] = get_allowed_hosts()

    processors = processors or []
    if not config.is_streams_service:
        processors = [
            "django.contrib.auth.context_processors.auth",
            "django.contrib.messages.context_processors.messages",
        ] + processors
    processors = [
        "django.template.context_processors.debug",
        "django.template.context_processors.request",
        "django.template.context_processors.media",
        "django.template.context_processors.static",
        "django.template.context_processors.tz",
        "haupt.common.settings.context_processors.version",
        "haupt.common.settings.context_processors.ui_assets_version",
        "haupt.common.settings.context_processors.ui_base_url",
        "haupt.common.settings.context_processors.ui_offline",
        "haupt.common.settings.context_processors.ui_enabled",
        "haupt.common.settings.context_processors.ui_in_sandbox",
    ] + processors

    context["FRONTEND_DEBUG"] = config.frontend_debug

    template_debug = config.template_debug or config.is_debug_mode
    context["UI_ADMIN_ENABLED"] = config.ui_admin_enabled
    base_url = config.ui_base_url
    if base_url:
        context["UI_BASE_URL"] = base_url
        context["FORCE_SCRIPT_NAME"] = base_url
    else:
        context["UI_BASE_URL"] = "/"
    context["UI_ASSETS_VERSION"] = config.ui_assets_version
    context["UI_OFFLINE"] = config.ui_offline
    context["UI_ENABLED"] = config.ui_enabled
    context["UI_IN_SANDBOX"] = config.ui_in_sandbox
    context["TEMPLATES_DEBUG"] = template_debug
    context["LIST_TEMPLATE_CONTEXT_PROCESSORS"] = processors
    context["TEMPLATES"] = [
        {
            "BACKEND": "django.template.backends.django.DjangoTemplates",
            "APP_DIRS": True,
            "OPTIONS": {"debug": template_debug, "context_processors": processors},
        }
    ]



from django.core.validators import validate_slug
from django.db import models

from haupt.db.abstracts.contributors import ContributorsModel
from haupt.db.abstracts.diff import DiffModel
from haupt.db.abstracts.duration import DurationModel
from haupt.db.abstracts.nameable import NameableModel
from haupt.db.abstracts.status import StatusModel
from haupt.db.abstracts.tag import TagModel
from haupt.db.abstracts.uid import UuidModel
from haupt.db.defs import Models


class BaseEvent(
    UuidModel,
    DiffModel,
    DurationModel,
    NameableModel,
    StatusModel,
    TagModel,
    ContributorsModel,
):
    name = models.CharField(
        max_length=128,
        blank=True,
        null=True,
        default=None,
        validators=[validate_slug],
    )
    version = models.ForeignKey(
        "db.ProjectVersion", on_delete=models.CASCADE, related_name="events"
    )
    session = models.CharField(
        max_length=128,
        blank=True,
        null=True,
        default=None,
        db_index=True,
        validators=[validate_slug],
    )
    user = models.CharField(
        max_length=128,
        blank=True,
        null=True,
        default=None,
        db_index=True,
        validators=[validate_slug],
    )
    project = models.ForeignKey(
        Models.get_db_model_name("Project"),
        on_delete=models.CASCADE,
        related_name="runs",
    )
    meta_info = models.JSONField(null=True, blank=True, default=dict)
    inputs = models.JSONField(null=True, blank=True)
    outputs = models.JSONField(null=True, blank=True)
    parent = models.ForeignKey(
        "self",
        on_delete=models.CASCADE,
        related_name="children",
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True
        app_label = "db"
        db_table = "db_event"
        indexes = [models.Index(fields=["name"])]



from django.db import models

from haupt.common.validation.slugs import validate_slug_with_dots
from haupt.db.abstracts.catalogs import BaseCatalog
from haupt.db.abstracts.contributors import ContributorsModel
from haupt.db.abstracts.readme import ReadmeModel
from haupt.db.abstracts.stage import StageModel
from haupt.db.abstracts.state import OptionalStateModel
from polyaxon.schemas import V1ProjectVersionKind


class BaseProjectVersion(
    BaseCatalog, StageModel, ReadmeModel, OptionalStateModel, ContributorsModel
):
    live_state = None

    kind = models.CharField(
        max_length=12,
        db_index=True,
        choices=V1ProjectVersionKind.to_choices(),
    )
    name = models.CharField(max_length=128, validators=[validate_slug_with_dots])
    project = models.ForeignKey(
        "db.Project", on_delete=models.CASCADE, related_name="versions"
    )
    content = models.TextField(
        help_text="The yaml/json content/metadata.",
        blank=True,
        null=True,
    )
    lineage = models.ManyToManyField(
        "db.ArtifactLineage", blank=True, related_name="versions"
    )
    run = models.ForeignKey(
        "db.Run",
        on_delete=models.CASCADE,
        related_name="versions",
        blank=True,
        null=True,
    )

    class Meta(BaseCatalog.Meta):
        abstract = True
        app_label = "db"
        db_table = "db_projectversion"
        unique_together = (("project", "name", "kind"),)



from django.core.validators import RegexValidator
from django.utils.regex_helper import _lazy_re_compile

try:
    from django.utils.safestring import mark_safe
except ImportError:
    raise ImportError("This module depends on django.")


from clipped.utils.strings import slugify as core_slugify


def slugify(value: str) -> str:
    return core_slugify(value, mark_safe)


slug_dots_re = _lazy_re_compile(r"^[-a-zA-Z0-9_.]+\Z")
validate_slug_with_dots = RegexValidator(
    slug_dots_re,
    # Translators: "letters" means latin letters: a-z and A-Z.
    "Enter a valid “value” consisting of letters, numbers, underscores, hyphens, or dots.",
    "invalid",
)



from haupt.common.query.service import QueryService
from haupt.common.service_interface import LazyServiceWrapper

backend = LazyServiceWrapper(
    backend_base=QueryService,
    backend_path="haupt.common.query.service.QueryService",
    options={},
)
backend.expose(locals())



from clipped.utils.imports import import_string

from haupt.common.events.event_service import EventService
from haupt.common.events.registry import run
from haupt.orchestration.executor.manager import event_manager


class ExecutorService(EventService):
    event_manager = event_manager

    def __init__(self, workers_service=None, handlers_service=None):
        self.workers_service = workers_service
        self.handlers_service = handlers_service
        self.workers = None
        self.handlers = None
        self.handlers_mapping = None

    def record_event(self, event: "Event") -> None:  # noqa: F821
        if self.workers and event.event_type in self.handlers_mapping:
            self.handlers_mapping[event.event_type](
                workers_backend=self.workers, event=event
            )

    @staticmethod
    def _get_handler_mapping(handlers):
        return {
            run.RUN_CREATED: handlers.handle_run_created,
            run.RUN_RESUMED_ACTOR: handlers.handle_run_created,
            run.RUN_APPROVED_ACTOR: handlers.handle_run_approved_triggered,
            run.RUN_STOPPED_ACTOR: handlers.handle_run_stopped_triggered,
            run.RUN_SKIPPED_ACTOR: handlers.handle_run_skipped_triggered,
            run.RUN_NEW_ARTIFACTS: handlers.handle_new_artifacts,
            run.RUN_DELETED_ACTOR: handlers.handle_run_deleted,
            run.RUN_NEW_STATUS: handlers.handle_run_new_status,
            run.RUN_DONE: handlers.handle_run_done,
        }

    def setup(self) -> None:
        super().setup()
        # Load default event types
        import haupt.orchestration.executor.subscriptions  # noqa

        if self.workers_service:
            self.workers = import_string(self.workers_service)
        if self.handlers_service:
            self.handlers = import_string(self.handlers_service)
            self.handlers_mapping = self._get_handler_mapping(self.handlers)



from django.db.models import F, Q

from haupt.background.celeryp.tasks import SchedulerCeleryTasks
from haupt.common import workers
from haupt.db.defs import Models


class CronsStatsManager:
    @staticmethod
    def stats_calculation_projects():
        projects = Models.Project.objects.filter(
            Q(updated_at__gt=F("latest_stats__updated_at"))
            | Q(latest_stats__isnull=True)
        )
        project_ids = set(projects.values_list("id", flat=True))

        if not project_ids:
            return

        for project_id in project_ids:
            workers.send(
                SchedulerCeleryTasks.STATS_CALCULATION_PROJECT,
                kwargs={"project_id": project_id},
            )
        return project_ids



from django.core.management import BaseCommand


class Command(BaseCommand):
    def handle(self, *args, **options):
        from haupt.common import workers
        from kombu import Exchange

        Exchange(
            "internal", type="topic", channel=workers.app.connection().channel()
        ).declare()



from datetime import datetime
from typing import List, Optional

from django.conf import settings
from django.db.models import Q
from django.utils.timezone import now

from haupt.common.authentication.base import is_normal_user
from haupt.db.defs import Models


def update_project_based_on_last_updated_entities(
    last_created_at_threshold: datetime,
    last_updated_at_threshold: datetime,
):
    current_check = now()
    project_ids_with_recent_runs = (
        Models.Run.all.filter(
            project__updated_at__lt=last_updated_at_threshold,
        )
        .filter(
            Q(created_at__gte=last_created_at_threshold)
            | Q(updated_at__gte=last_created_at_threshold)
        )
        .values_list("project_id", flat=True)
        .distinct()
    )
    project_ids_with_recent_versions = (
        Models.ProjectVersion.objects.filter(
            project__updated_at__lt=last_updated_at_threshold,
            created_at__gte=last_created_at_threshold,
        )
        .values_list("project_id", flat=True)
        .distinct()
    )

    project_ids_with_recent_entities = set(project_ids_with_recent_runs) | set(
        project_ids_with_recent_versions
    )
    if not project_ids_with_recent_entities:
        return

    Models.Project.objects.filter(id__in=project_ids_with_recent_entities).update(
        updated_at=current_check
    )


def add_project_contributors(
    project: Models.Project,
    users: Optional[List[Models.User]] = None,
    user_ids: Optional[List[int]] = None,
):
    if not settings.HAS_ORG_MANAGEMENT:
        return
    if not project:
        return
    _users = [u.id for u in users if is_normal_user(u)] if users else user_ids
    if not _users:
        return

    project.contributors.add(*_users)



from django.db import models


class DurationModel(models.Model):
    started_at = models.DateTimeField(blank=True, null=True)
    finished_at = models.DateTimeField(blank=True, null=True)
    wait_time = models.FloatField(blank=True, null=True)
    duration = models.FloatField(blank=True, null=True)

    class Meta:
        abstract = True



from typing import Dict

from clipped.utils.json import orjson_dumps

from polyaxon._fs.async_manager import download_file, upload_data
from polyaxon._fs.types import FSSystem


async def upload_op_spec(fs: FSSystem, store_path: str, run_uuid: str, op_spec: Dict):
    subpath = "{}/outputs/spec.json".format(run_uuid)
    await upload_data(
        fs=fs, store_path=store_path, subpath=subpath, data=orjson_dumps(op_spec)
    )


async def download_op_spec(fs: FSSystem, store_path: str, run_uuid: str):
    subpath = "{}/outputs/spec.json".format(run_uuid)
    return await download_file(
        fs=fs,
        store_path=store_path,
        subpath=subpath,
    )


async def upload_agent_spec(
    fs: FSSystem, store_path: str, agent_uuid: str, agent_spec: Dict
):
    subpath = ".agents/{}/outputs/spec.json".format(agent_uuid)
    await upload_data(
        fs=fs, store_path=store_path, subpath=subpath, data=orjson_dumps(agent_spec)
    )


async def download_agent_spec(fs: FSSystem, store_path: str, agent_uuid: str):
    subpath = ".agents/{}/outputs/spec.json".format(agent_uuid)
    return await download_file(
        fs=fs,
        store_path=store_path,
        subpath=subpath,
    )



from django.db import models

from haupt.db.abstracts.describable import DescribableModel
from haupt.db.abstracts.diff import DiffModel
from haupt.db.abstracts.live_state import LiveStateModel
from haupt.db.abstracts.nameable import RequiredNameableModel
from haupt.db.abstracts.tag import TagModel
from haupt.db.abstracts.uid import UuidModel


class BaseCatalog(
    UuidModel,
    RequiredNameableModel,
    DiffModel,
    DescribableModel,
    TagModel,
):
    class Meta:
        abstract = True
        indexes = [models.Index(fields=["name"])]


class BaseLiveStateCatalog(
    BaseCatalog,
    LiveStateModel,
):
    class Meta:
        abstract = True
        indexes = [models.Index(fields=["name"])]



from typing import Optional


def get_num_to_start(concurrency: int, consumed: int, max_budget: Optional[int]) -> int:
    """We need to check if we are allowed to start the experiment
    If the polyaxonfile has concurrency we need to check how many experiments are running.

    N.B. max_budget is set ~ inf
    """
    if concurrency is None or concurrency < 0:  # We max all runs
        concurrency = max_budget
        max_budget = None

    if concurrency is None:
        return 0

    if max_budget is None:  # No budget only check concurrency
        # Target to start: should try to always schedule to the target concurrency
        return concurrency - consumed

    if max_budget <= 0:  # Not allowed to start anything
        return 0

    if (
        max_budget and max_budget <= consumed
    ):  # Already consumed the max possible budget
        return 0

    # Either respect the concurrency settings or the allocatable budget if the budget is smaller
    max_to_start = max_budget - consumed

    # Target to start: should try to always schedule to the target concurrency
    target = concurrency - consumed

    return target if target < max_to_start else max_to_start



from haupt.common import auditor
from haupt.common.events.registry import model_version

auditor.subscribe(model_version.ModelVersionDeletedEvent)
auditor.subscribe(model_version.ModelVersionCreatedActorEvent)
auditor.subscribe(model_version.ModelVersionUpdatedActorEvent)
auditor.subscribe(model_version.ModelVersionViewedActorEvent)
auditor.subscribe(model_version.ModelVersionDeletedActorEvent)
auditor.subscribe(model_version.ModelVersionTransferredActorEvent)
auditor.subscribe(model_version.ModelVersionNewStageEvent)



from typing import Any, Optional

from haupt.common.conf.conf_manager import conf_cache_manager
from haupt.common.conf.exceptions import ConfException
from haupt.common.conf.option_manager import option_manager
from haupt.common.options.option import Option
from haupt.common.options.option_owners import OptionOwners
from haupt.common.service_interface import Service
from polyaxon.exceptions import PolyaxonException


class OptionService(Service):
    __all__ = ("get", "set", "delete", "clear_cache", "get_owners")

    option_manager = option_manager
    cache_manager = conf_cache_manager
    service_name = "Option"

    def __init__(self, check_ownership=False):
        self.stores = {}
        self.check_ownership = check_ownership

    def get_options_handler(self):
        return None

    def can_handle(self, key: str) -> bool:
        return isinstance(key, str) and self.option_manager.knows(key=key)

    def get_option(self, key: str, owners: OptionOwners) -> Option:
        option = self.option_manager.get(key=key)
        if self.check_ownership and not option.is_global():
            if not owners or not [i for i in owners if i]:
                raise ConfException("Option `{}` requires an owner.".format(option.key))
        return option

    def get_store(self, option: Option) -> Any:
        if option.store not in self.stores:
            raise ConfException("Option `{}` has an invalid store.".format(option.key))

        return self.stores[option.store]

    @staticmethod
    def _get_value(option, value, to_dict):
        if not to_dict:
            return value
        option_dict = option.to_dict(value=value)
        return option_dict

    def get(
        self,
        key: str,
        check_cache: bool = True,
        to_dict: bool = False,
        owners: Optional[OptionOwners] = None,
    ) -> Any:
        if not self.is_setup:
            return
        if not self.can_handle(key=key):
            raise ConfException(
                "{} service request an unknown key `{}`.".format(self.service_name, key)
            )

        option = self.get_option(key=key, owners=owners)

        if check_cache:
            value = self.cache_manager.get_from_cache(key=key, owners=owners)
            if self.cache_manager.is_valid_value(value=value):
                return self._get_value(option=option, value=value, to_dict=to_dict)

        store = self.get_store(option=option)
        value = store.get(option=option, owners=owners)

        # Cache value
        self.cache_manager.set_to_cache(
            key=key, value=value, ttl=option.cache_ttl, owners=owners
        )

        return self._get_value(option=option, value=value, to_dict=to_dict)

    def set(self, key: str, value: Any, owners: Optional[OptionOwners] = None):
        if not self.is_setup:
            return
        if not self.can_handle(key=key):
            raise ConfException(
                "{} service request an unknown key `{}`.".format(self.service_name, key)
            )
        if value is None:
            raise ConfException(
                "{} service requires a value for key `{}` to set.".format(
                    self.service_name, key
                )
            )
        option = self.get_option(key=key, owners=owners)
        # Convert value
        try:
            value = option.parse(value=value)
        except PolyaxonException as e:
            raise ConfException(e)

        store = self.get_store(option=option)
        store.set(option=option, value=value, owners=owners)
        # Cache value
        self.cache_manager.set_to_cache(
            key=key, value=value, ttl=option.cache_ttl, owners=owners
        )

    def delete(self, key: str, owners: Optional[OptionOwners] = None):
        if not self.is_setup:
            return
        if not self.can_handle(key=key):
            raise ConfException(
                "{} service request an unknown key `{}`.".format(self.service_name, key)
            )

        option = self.get_option(key=key, owners=owners)
        store = self.get_store(option=option)
        store.delete(option=option, owners=owners)
        # Clear Cache key
        self.cache_manager.clear_key(key=key, owners=owners)

    def clear_cache(self):
        self.cache_manager.clear()

    @staticmethod
    def get_owners(
        user: Optional[int] = None,
        project: Optional[int] = None,
        organization: Optional[int] = None,
    ) -> OptionOwners:
        return OptionOwners.get_owners(
            user=user, project=project, organization=organization
        )



from clipped.decorators.signals import ignore_raw

from django.db.models.signals import post_delete
from django.dispatch import receiver

from haupt.common.content_types import ContentTypes
from haupt.db.defs import Models
from haupt.db.managers.bookmarks import remove_bookmarks


@receiver(post_delete, sender=Models.Run, dispatch_uid="run_bookmark_post_delete")
@ignore_raw
def run_bookmark_post_delete(sender, **kwargs):
    instance = kwargs["instance"]
    remove_bookmarks(object_id=instance.id, content_type=ContentTypes.RUN.value)


@receiver(
    post_delete, sender=Models.Project, dispatch_uid="project_bookmark_post_delete"
)
@ignore_raw
def project_bookmark_post_delete(sender, **kwargs):
    instance = kwargs["instance"]
    remove_bookmarks(object_id=instance.id, content_type=ContentTypes.PROJECT.value)



from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from polyaxon.schemas import ManagedBy


class IsManagedMixin(serializers.Serializer):
    is_managed = serializers.BooleanField(initial=True, default=True, allow_null=True)

    def _get_is_managed(self, value):
        return value if isinstance(value, bool) else True

    def check_if_entity_is_managed(self, attrs, entity_name, config_field="content"):
        is_managed = None
        managed_by = None
        if "managed_by" in attrs:
            managed_by = attrs.get("managed_by")
            if not managed_by:
                managed_by = ManagedBy.AGENT
            if managed_by == ManagedBy.USER:
                is_managed = False
            elif ManagedBy.is_managed(managed_by):
                is_managed = True
        elif "is_managed" in attrs:
            is_managed = self._get_is_managed(attrs.get("is_managed"))
            if is_managed:
                managed_by = ManagedBy.AGENT
            else:
                managed_by = ManagedBy.USER

        if is_managed is None and managed_by is None:
            return attrs
        if is_managed and not attrs.get(config_field):
            raise ValidationError(
                "{} expects a `{}`.".format(entity_name, config_field)
            )
        attrs["managed_by"] = managed_by
        attrs.pop("is_managed", None)
        return attrs

    def validate_is_managed(self, value):
        return self._get_is_managed(value)



from typing import Dict, List

from django.conf import settings

from haupt.db.defs import Models

artifacts_names = Models.ArtifactLineage.objects.prefetch_related("artifact").only(
    "artifact__id", "artifact__name"
)

artifacts = (
    Models.ArtifactLineage.objects.prefetch_related("artifact")
    .only(
        "is_input",
        "artifact__id",
        "artifact__name",
        "artifact__kind",
        "artifact__path",
        "artifact__state",
        "artifact__summary",
    )
    .distinct()
)

project_runs_artifacts = (
    Models.ArtifactLineage.objects.prefetch_related("artifact", "run")
    .only(
        "is_input",
        "artifact__id",
        "artifact__name",
        "artifact__kind",
        "artifact__path",
        "artifact__state",
        "artifact__summary",
        "run__id",
        "run__uuid",
        "run__name",
    )
    .distinct()
)

project_runs_artifacts_distinct = Models.ArtifactLineage.objects.prefetch_related(
    "artifact"
).only(
    "is_input",
    "artifact__name",
    "artifact__kind",
)
if settings.DB_ENGINE_NAME == "sqlite":
    project_runs_artifacts_distinct = project_runs_artifacts_distinct.distinct()
else:
    project_runs_artifacts_distinct = project_runs_artifacts_distinct.distinct(
        "is_input",
        "artifact__name",
        "artifact__kind",
    )


def clean_sqlite_distinct_artifacts(data: List[Dict]):
    """Sqlite does not provide a distinct logic based on columns"""
    distinct_data = []
    data_ids = set([])
    count = 0
    for d in data:
        d_id = "{}-{}-{}".format(d["is_input"], d["name"], d["kind"])
        if d_id not in data_ids:
            data_ids.add(d_id)
            distinct_data.append(d)
            count += 1

    return distinct_data, count



from haupt.common.conf.handlers.env_handler import EnvConfHandler
from haupt.common.conf.handlers.settings_handler import SettingsConfHandler
from haupt.common.conf.option_service import OptionService
from haupt.common.options.option import OptionStores


class ConfService(OptionService):
    def setup(self):
        super().setup()
        self.stores[OptionStores.SETTINGS] = SettingsConfHandler()
        self.stores[OptionStores.ENV] = EnvConfHandler()



import uuid

from typing import List, Optional

from clipped.utils.lists import to_list

from haupt.db.defs import Models
from polyaxon._k8s.k8s_schemas import V1Container
from polyaxon.schemas import V1IO, V1Cache, V1CloningKind, V1Init


def get_run_io_state(
    run_state: str,
    cache: V1Cache,
    inputs: List[V1IO],
    outputs: List[V1IO],
    contexts: List[V1IO],
) -> str:
    cache_io = []
    if cache and cache.io:
        cache_io = cache.io or []

    def check_cache_io(io: V1IO, check_value: bool = False):
        if check_value and io.value is None:
            return False
        if cache_io:
            return io.name in cache_io
        return True

    inputs = sorted(inputs or [], key=lambda i: i.name)
    io_reprs = [io.get_repr() for io in inputs if check_cache_io(io)]
    if io_reprs:
        run_state += "in-" + str(io_reprs)

    outputs = sorted(outputs or [], key=lambda i: i.name)
    io_reprs = [io.get_repr() for io in outputs if check_cache_io(io, True)]
    if io_reprs:
        run_state += "out-" + str(io_reprs)

    contexts = sorted(contexts or [], key=lambda i: i.name)
    io_reprs = [io.get_repr() for io in contexts if check_cache_io(io, True)]
    if io_reprs:
        run_state += "ctx-" + str(io_reprs)

    return run_state


def get_state_from_container(container: V1Container):
    if not container:
        return None

    result = ""
    if container.command:
        result += "cmd-" + str(
            [str(c) for c in to_list(container.command, check_none=True) if c]
        )
    if container.args:
        result += "args-" + str(
            [str(c) for c in to_list(container.args, check_none=True) if c]
        )

    return result


def get_run_containers_state(
    containers: List[V1Container],
) -> Optional[str]:
    containers = containers or []
    cache_containers = [get_state_from_container(c) for c in containers]
    cache_containers = [c for c in cache_containers if c]
    if cache_containers:
        return str(cache_containers)

    return None


def get_run_connections_state(
    run_state: str,
    connections: List[str],
) -> str:
    connections = connections or []
    connections = [c for c in connections if c]
    if connections:
        return run_state + "connections-" + str(connections)

    return run_state


def get_run_init_state(
    run_state: str,
    init: List[V1Init],
) -> str:
    init = init or []
    init_reprs = [i.to_light_dict(exclude_attrs=["container"]) for i in init if i]
    if init_reprs:
        init_reprs += "init-" + str(init_reprs)

    container_reprs = get_run_containers_state(
        [i.container for i in init if i.container]
    )
    if container_reprs:
        return run_state + "init-containers-" + str(container_reprs)

    return run_state


def get_run_sections_state(
    run_state: str,
    cache: V1Cache,
    init: List[V1Init],
    connections: List[str],
    containers: List[V1Container],
) -> str:
    if cache and cache.sections is not None:
        cache_sections = cache.sections
    else:
        cache_sections = ["init", "connections", "containers"]

    if "init" in cache_sections and init:
        run_state += get_run_init_state(run_state, init)

    if "connections" in cache_sections and connections:
        run_state += get_run_connections_state(run_state, connections)

    if "containers" in cache_sections and containers:
        containers_reprs = get_run_containers_state(containers)
        if containers_reprs:
            run_state += "containers-" + str(containers_reprs)

    return run_state


def get_run_state(
    cache: V1Cache,
    inputs: List[V1IO],
    outputs: List[V1IO],
    contexts: List[V1IO],
    init: List[V1Init],
    connections: List[str],
    containers: List[V1Container],
    namespace: uuid.UUID,
    component_state: uuid.UUID = None,
) -> Optional[uuid.UUID]:
    """A string representation that is used to create hash cache"""
    run_state = ""

    run_state = get_run_io_state(
        run_state=run_state,
        cache=cache,
        inputs=inputs,
        outputs=outputs,
        contexts=contexts,
    )

    run_state = get_run_sections_state(
        run_state=run_state,
        cache=cache,
        init=init,
        connections=connections,
        containers=containers,
    )

    if not run_state:
        return None

    if component_state:
        run_state = component_state.hex + run_state
    return uuid.uuid5(namespace, run_state)


def get_cache_clones(run_id: int):
    return Models.Run.objects.filter(
        original_id=run_id, cloning_kind=V1CloningKind.CACHE
    ).values_list("id", flat=True)



import logging

from typing import Optional

from clipped.compact.pydantic import Field, StrictInt, StrictStr, root_validator
from clipped.utils.http import clean_host
from clipped.utils.paths import check_or_create_path

from polyaxon._env_vars.keys import (
    ENV_KEYS_K8S_NAMESPACE,
    ENV_KEYS_SANDBOX_DEBUG,
    ENV_KEYS_SANDBOX_HOST,
    ENV_KEYS_SANDBOX_PER_CORE,
    ENV_KEYS_SANDBOX_PORT,
    ENV_KEYS_SANDBOX_SSL_ENABLED,
    ENV_KEYS_SANDBOX_WORKERS,
    ENV_KEYS_SERVICE_MODE,
)
from polyaxon._schemas.agent import AgentConfig
from polyaxon._services import PolyaxonServices

_logger = logging.getLogger("sandbox.config")


class SandboxConfig(AgentConfig):
    _IDENTIFIER = "sandbox"

    namespace: Optional[StrictStr] = Field(
        alias=ENV_KEYS_K8S_NAMESPACE, default="sandbox"
    )
    port: Optional[StrictInt] = Field(alias=ENV_KEYS_SANDBOX_PORT)
    host: Optional[StrictStr] = Field(alias=ENV_KEYS_SANDBOX_HOST)
    ssl_enabled: Optional[bool] = Field(alias=ENV_KEYS_SANDBOX_SSL_ENABLED)
    debug: Optional[bool] = Field(alias=ENV_KEYS_SANDBOX_DEBUG)
    workers: Optional[StrictInt] = Field(alias=ENV_KEYS_SANDBOX_WORKERS)
    per_core: Optional[bool] = Field(alias=ENV_KEYS_SANDBOX_PER_CORE)
    mode: Optional[StrictStr] = Field(alias=ENV_KEYS_SERVICE_MODE)

    def __init__(
        self,
        host: Optional[str] = None,
        **data,
    ):
        super().__init__(
            host=clean_host(host) if host else host,
            **data,
        )

    @root_validator(pre=True)
    def handle_camel_case_sandbox(cls, values):
        if (
            not values.get("ssl_enabled")
            and not values.get(ENV_KEYS_SANDBOX_SSL_ENABLED)
            and "sslEnabled" in values
        ):
            values[ENV_KEYS_SANDBOX_SSL_ENABLED] = values["sslEnabled"]
        if (
            not values.get("per_core")
            and not values.get(ENV_KEYS_SANDBOX_PER_CORE)
            and "perCore" in values
        ):
            values[ENV_KEYS_SANDBOX_PER_CORE] = values["perCore"]
        return values

    def mount_sandbox(self, path: Optional[str] = None):
        from polyaxon._contexts.paths import mount_sandbox

        if not path:
            return

        mount_sandbox(path=path)
        try:
            check_or_create_path(path, is_dir=True)
        except Exception as e:
            _logger.error("Could not create sandbox path `%s`.", path)
            raise e

    @property
    def is_viewer(self):
        return self.mode == PolyaxonServices.VIEWER



import logging
import os

from typing import Optional

from clipped.utils.workers import get_core_workers

import uvicorn

from haupt.cli.runners.cron import start_cron
from haupt.cli.runners.manage import migrate
from polyaxon._env_vars.keys import ENV_KEYS_PROXY_LOCAL_PORT

_logger = logging.getLogger("haupt.cli")


def start_app(
    app,
    app_name,
    host: Optional[str] = None,
    port: Optional[int] = None,
    log_level: Optional[str] = None,
    workers: Optional[int] = None,
    per_core: bool = False,
    max_workers: Optional[int] = None,
    uds: Optional[str] = None,
    migrate_tables: bool = False,
    migrate_db: bool = False,
    use_cron: bool = False,
):
    migrate(migrate_tables=migrate_tables, migrate_db=migrate_db)
    host = host or "0.0.0.0"
    port = int(port or 8000)
    os.environ[ENV_KEYS_PROXY_LOCAL_PORT] = str(port)
    log_level = log_level or "warning"
    if per_core:
        workers = get_core_workers(per_core=workers or 2, max_workers=max_workers)
    else:
        workers = workers or get_core_workers(per_core=2, max_workers=max_workers)

    dashboard_url = os.environ.get("POLYAXON_DASHBOARD_URL", f"http://{host}:{port}")
    _logger.info(
        "{app_name} is running on {dashboard_url} in process {pid} with {workers} workers".format(
            app_name=app_name,
            dashboard_url=dashboard_url,
            pid=os.getpid(),
            workers=workers,
        )
    )
    if use_cron:
        start_cron(host=f"http://{host}:{port}")
    uvicorn.run(
        app,
        host=host,
        port=port,
        access_log=False,
        log_level=log_level.lower(),
        workers=int(workers),
        uds=uds,
        timeout_keep_alive=120,
    )



import logging

from datetime import datetime
from functools import reduce
from operator import __or__ as OR
from typing import Dict, List, Optional

from clipped.compact.pydantic import ValidationError as PydanticValidationError
from clipped.utils.lists import to_list
from rest_framework.exceptions import ValidationError

from django.db.models import Q, QuerySet
from django.utils.timezone import now

from haupt.background.celeryp.tasks import SchedulerCeleryTasks
from haupt.common import workers
from haupt.common.exceptions import AccessNotAuthorized, AccessNotFound
from haupt.db.defs import Models
from haupt.db.managers import flows
from haupt.db.managers.artifacts import atomic_set_artifacts
from haupt.db.managers.live_state import (
    delete_in_progress_project,
    delete_in_progress_run,
)
from haupt.db.managers.runs import (
    base_approve_run,
    get_failed_stopped_and_all_runs,
    get_stopping_pipelines_with_no_runs,
    is_pipeline_done,
)
from haupt.db.managers.stats import (
    collect_project_run_count_stats,
    collect_project_run_duration_stats,
    collect_project_run_status_stats,
    collect_project_unique_user_stats,
    collect_project_version_stats,
)
from haupt.db.managers.statuses import (
    bulk_new_run_status,
    new_run_skip_status,
    new_run_status,
    new_run_stop_status,
)
from haupt.db.queries.runs import (
    STATUS_UPDATE_COLUMNS_DEFER,
    STATUS_UPDATE_COLUMNS_ONLY,
)
from haupt.orchestration import operations
from haupt.orchestration.scheduler.resolver import SchedulingResolver
from polyaxon._compiler import resolver
from polyaxon._constants.metadata import (
    META_BRACKET_ITERATION,
    META_DESTINATION_IMAGE,
    META_HAS_EARLY_STOPPING,
    META_HAS_HOOKS,
    META_IS_HOOK,
    META_ITERATION,
)
from polyaxon._operations import get_bo_tuner, get_hyperband_tuner, get_hyperopt_tuner
from polyaxon.exceptions import (
    PolyaxonCompilerError,
    PolyaxonException,
    PolyaxonSchemaError,
)
from polyaxon.schemas import (
    LifeCycle,
    ManagedBy,
    V1CloningKind,
    V1CompiledOperation,
    V1FailureEarlyStopping,
    V1MatrixKind,
    V1MetricEarlyStopping,
    V1Operation,
    V1Optimization,
    V1RunEdgeKind,
    V1RunKind,
    V1RunPending,
    V1StatusCondition,
    V1Statuses,
    V1TriggerPolicy,
    dags,
)
from traceml.artifacts import V1ArtifactKind, V1RunArtifact

_logger = logging.getLogger("polyaxon.scheduler")


class SchedulingManager:
    DEFAULT_PREFETCH = ["project"]
    RESOLVER = SchedulingResolver

    @classmethod
    def _resolve(
        cls,
        run: Models.Run,
        compiled_at: Optional[datetime] = None,
    ):
        try:
            compiled_operation = V1CompiledOperation.read(
                run.content
            )  # TODO: Use construct
            project = run.project
            return resolver.resolve(
                run=run,
                compiled_operation=compiled_operation,
                owner_name=project.owner.name,
                project_name=project.name,
                project_uuid=project.uuid.hex,
                run_uuid=run.uuid.hex,
                run_name=run.name,
                run_path=run.subpath,
                resolver_cls=cls.RESOLVER,
                params=None,
                compiled_at=compiled_at,
                created_at=run.created_at,
                cloning_kind=run.cloning_kind,
                original_uuid=run.original.uuid.hex if run.original_id else None,
                is_independent=not bool(run.pipeline_id),
            )
        except (
            AccessNotAuthorized,
            AccessNotFound,
        ) as e:
            raise PolyaxonCompilerError("Access Error: %s" % e) from e
        except (
            AccessNotAuthorized,
            AccessNotFound,
            PydanticValidationError,
            PolyaxonSchemaError,
            ValidationError,
        ) as e:
            raise PolyaxonCompilerError("Compilation Error: %s" % e) from e

    @classmethod
    def _resolve_hooks(cls, run: Models.Run) -> List[V1Operation]:
        try:
            compiled_operation = V1CompiledOperation.read(
                run.content
            )  # TODO: Use construct
            project = run.project
            return resolver.resolve_hooks(
                run=run,
                compiled_operation=compiled_operation,
                owner_name=project.owner.name,
                project_name=project.name,
                project_uuid=project.uuid.hex,
                run_uuid=run.uuid.hex,
                run_name=run.name,
                run_path=run.subpath,
                resolver_cls=cls.RESOLVER,
                params=None,
                compiled_at=None,
                created_at=run.created_at,
                cloning_kind=run.cloning_kind,
                original_uuid=run.original.uuid.hex if run.original_id else None,
            )
        except (
            AccessNotAuthorized,
            AccessNotFound,
        ) as e:
            raise PolyaxonCompilerError("Access Error: %s" % e) from e
        except (
            AccessNotAuthorized,
            AccessNotFound,
            PydanticValidationError,
            PolyaxonSchemaError,
            ValidationError,
        ) as e:
            raise PolyaxonCompilerError("Compilation Error: %s" % e) from e

    @staticmethod
    def _capture_exception(e: Exception):
        _logger.exception(e)

    @staticmethod
    def _capture_message(message: str):
        return

    @staticmethod
    def _set_iteration_meta_info(
        run: Models.Run, suggestions: List[Dict], iteration: int, bracket_iteration: int
    ) -> V1CompiledOperation:
        compiled_operation = V1CompiledOperation.read(
            run.content
        )  # TODO: Use construct

        condition = not suggestions or iteration is None
        if condition or (
            compiled_operation.has_hyperband_matrix and bracket_iteration is None
        ):
            _logger.info(
                "Run `%s` cannot iterate and create matrix runs, received wrong iteration lineage",
                run.id,
            )
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.FAILED,
                status="True",
                reason="SchedulerIterateFailure",
                message="Run received wrong iteration lineage, "
                "an iteration number and a list of suggestions is required, "
                "received empty suggestions.",
            )
            new_run_status(run=run, condition=condition)
            return compiled_operation
        run.meta_info[META_ITERATION] = iteration
        if compiled_operation.has_hyperband_matrix:
            run.meta_info[META_BRACKET_ITERATION] = bracket_iteration
        run.save(update_fields=["meta_info"])
        return compiled_operation

    @classmethod
    def _execute_hooks(
        cls, run: Models.Run, hook_ops: List[V1Operation]
    ) -> List[Models.Run]:
        if not hook_ops:
            return []

        hub_refs_to_components = cls.RESOLVER._collect_hub_refs(
            ops=hook_ops, owner_name=run.project.owner.name
        )

        def get_edges():
            run_edges = []
            for hook_run in runs:
                run_edges.append(
                    Models.RunEdge(
                        upstream_id=run.id,
                        downstream_id=hook_run.id,
                        kind=V1RunEdgeKind.HOOK,
                    )
                )
            return run_edges

        runs = []
        for hook in hook_ops:
            component_state = None
            if (
                not hook.has_component_reference
                and hook.has_hub_reference
                and hook.hub_ref in hub_refs_to_components
            ):
                hook.set_definition(
                    hub_refs_to_components[hook.hub_ref]["component_content"]
                )
                component_state = hub_refs_to_components[hook.hub_ref][
                    "component_state"
                ]
            hook_run = operations.init_run(
                project_id=run.project_id,
                user_id=run.user_id,
                op_spec=hook,
                supported_owners={run.project.owner.name},
                component_state=component_state,
            ).instance
            hook_run.meta_info[META_IS_HOOK] = True
            runs.append(hook_run)

        # Create batch runs
        if runs:
            runs = Models.Run.objects.bulk_create(runs)

        # Create edges
        run_edges = get_edges()
        if run_edges:
            Models.RunEdge.objects.bulk_create(run_edges)

        return [run.id for run in runs]

    @staticmethod
    def _notify_build(run: Models.Run):
        if run.runtime != V1RunKind.BUILDER:
            return
        runs = run.downstream_runs.exclude(status__in=LifeCycle.DONE_VALUES).filter(
            pending=V1RunPending.BUILD,
        )
        if run.status != V1Statuses.SUCCEEDED:
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.UPSTREAM_FAILED,
                status="True",
                reason="SchedulerBuildNotify",
                message=f"Build process {run.uuid.hex} did not succeed, "
                f"last build status: {run.status}.",
            )
            bulk_new_run_status(runs, condition)
            return

        destination = (
            Models.ArtifactLineage.objects.filter(
                artifact__name="destination",
                artifact__kind=V1ArtifactKind.DOCKER_IMAGE,
                is_input=False,
                run_id=run.id,
            )
            .values_list("artifact__summary", flat=True)
            .last()
        )
        destination_image = None
        if destination:
            destination_image = destination.get("image")

        if not destination_image:
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.UPSTREAM_FAILED,
                status="True",
                reason="SchedulerBuildNotify",
                message=f"Build process {run.uuid.hex} succeed "
                f"but did not generate an image destination.",
            )
            bulk_new_run_status(runs, condition)
            return

        built_runs = []
        for built_run in runs:
            meta_info = built_run.meta_info or {}
            meta_info[META_DESTINATION_IMAGE] = destination_image
            built_run.meta_info = meta_info
            built_runs.append(built_run)
        Models.Run.objects.bulk_update(built_runs, ["meta_info"])
        for built_run in built_runs:
            workers.send(
                SchedulerCeleryTasks.RUNS_BUILT,
                kwargs={"run_id": built_run.id},
                eager_kwargs={"run": built_run},
            )

    @staticmethod
    def _notify_cache(run: Models.Run):
        condition = V1StatusCondition.get_condition(
            type=run.status,
            status="True",
            reason="SchedulerCacheNotify",
            message="Original run is done.",
        )
        for clone_run in Models.Run.objects.filter(
            original_id=run.id,
            cloning_kind=V1CloningKind.CACHE,
            pending=V1RunPending.CACHE,
        ).only("outputs", *STATUS_UPDATE_COLUMNS_ONLY):
            clone_run.outputs = run.outputs
            clone_run.pending = None
            new_run_status(
                run=clone_run,
                condition=condition,
                additional_fields=["outputs", "pending"],
            )

    @staticmethod
    def _notify_hooks(run: Models.Run):
        if run.meta_info.get(META_HAS_HOOKS):
            workers.send(
                SchedulerCeleryTasks.RUNS_HOOKS,
                kwargs={"run_id": run.id},
                eager_kwargs={"run": run},
            )

    @staticmethod
    def _check_failure_early_stopping(
        run: Models.Run, early_stopping: List[V1FailureEarlyStopping]
    ) -> bool:
        # We only need one with the lowest percent
        early_stopping = to_list(early_stopping, check_none=True)
        if not early_stopping:
            return False
        percent = min([es.percent for es in early_stopping if es.percent])
        failed_count = run.controller_runs.filter(
            status__in={V1Statuses.FAILED, V1Statuses.UPSTREAM_FAILED}
        ).count()
        if failed_count == 0:
            return False
        return failed_count / run.controller_runs.count() >= (percent / 100)

    @staticmethod
    def _check_absolute_metric_early_stopping(
        run: Models.Run, early_stopping: List[V1MetricEarlyStopping]
    ) -> bool:
        early_stopping = to_list(early_stopping, check_none=True)
        if not early_stopping:
            return False
        filters = []
        for early_stopping_metric in early_stopping:
            comparison = (
                "gte"
                if V1Optimization.maximize(early_stopping_metric.optimization)
                else "lte"
            )
            metric_filter = "outputs__{}__{}".format(
                early_stopping_metric.metric, comparison
            )
            filters.append({metric_filter: early_stopping_metric.value})
        if filters:
            return run.controller_runs.filter(
                reduce(OR, [Q(**f) for f in filters])
            ).exists()
        return False

    @staticmethod
    def _check_metric_early_stopping(
        run: Models.Run, early_stopping: List[V1MetricEarlyStopping]
    ) -> bool:
        early_stopping = to_list(early_stopping, check_none=True)
        if not early_stopping:
            return False
        return False

    @classmethod
    def _should_early_stop(cls, run: Models.Run) -> bool:
        if not run.meta_info.get(META_HAS_EARLY_STOPPING):
            return False

        if hasattr(run, "_compiled_operation"):
            compiled_operation = run._compiled_operation
        else:
            compiled_operation = V1CompiledOperation.read(
                run.content
            )  # TODO: Use construct
            # Cache the compiled_operation to avoid recalculation
            run._compiled_operation = compiled_operation

        if compiled_operation.matrix:
            early_stopping = compiled_operation.matrix.early_stopping
        elif compiled_operation.is_dag_run:
            early_stopping = compiled_operation.run.early_stopping

        early_stopping = to_list(early_stopping, check_none=True)
        metric_early_stopping = []
        absolute_metric_early_stopping = []
        failure_early_stopping = []
        for es in early_stopping:
            if es.kind == V1FailureEarlyStopping._IDENTIFIER:
                failure_early_stopping.append(es)
            elif es.kind == V1MetricEarlyStopping._IDENTIFIER:
                if es.policy:
                    metric_early_stopping.append(es)
                else:
                    absolute_metric_early_stopping.append(es)

        if cls._check_failure_early_stopping(
            run, early_stopping=failure_early_stopping
        ):
            return True
        if cls._check_absolute_metric_early_stopping(
            run, early_stopping=absolute_metric_early_stopping
        ):
            return True
        if cls._check_metric_early_stopping(run, early_stopping=metric_early_stopping):
            return True

        return False

    @staticmethod
    def _get_upstream_runs(
        run: Models.Run, pipeline_run_id: Optional[int] = None
    ) -> QuerySet:
        upstream = run.upstream_runs
        if pipeline_run_id:
            upstream = upstream.filter(pipeline_id=pipeline_run_id)
        return upstream

    @classmethod
    def _is_upstream_all_done(
        cls, run: Models.Run, pipeline_run_id: Optional[int] = None
    ) -> bool:
        return (
            not cls._get_upstream_runs(run=run, pipeline_run_id=pipeline_run_id)
            .exclude(status__in=LifeCycle.DONE_VALUES)
            .exists()
        )

    @classmethod
    def _is_upstream_one_done(
        cls, run: Models.Run, pipeline_run_id: Optional[int] = None
    ) -> bool:
        return (
            cls._get_upstream_runs(run=run, pipeline_run_id=pipeline_run_id)
            .filter(status__in=LifeCycle.DONE_VALUES)
            .exists()
        )

    @classmethod
    def _is_upstream_all_failed(
        cls, run: Models.Run, pipeline_run_id: Optional[int] = None
    ) -> bool:
        return (
            not cls._get_upstream_runs(run=run, pipeline_run_id=pipeline_run_id)
            .exclude(status=V1Statuses.FAILED)
            .exists()
        )

    @classmethod
    def _is_upstream_one_failed(
        cls, run: Models.Run, pipeline_run_id: Optional[int] = None
    ) -> bool:
        return (
            cls._get_upstream_runs(run=run, pipeline_run_id=pipeline_run_id)
            .filter(status=V1Statuses.FAILED)
            .exists()
        )

    @classmethod
    def _is_upstream_all_succeeded(
        cls, run: Models.Run, pipeline_run_id: Optional[int] = None
    ) -> bool:
        return (
            not cls._get_upstream_runs(run=run, pipeline_run_id=pipeline_run_id)
            .exclude(status=V1Statuses.SUCCEEDED)
            .exists()
        )

    @classmethod
    def _is_upstream_one_succeeded(
        cls, run: Models.Run, pipeline_run_id: Optional[int] = None
    ) -> bool:
        return (
            cls._get_upstream_runs(run=run, pipeline_run_id=pipeline_run_id)
            .filter(status=V1Statuses.SUCCEEDED)
            .exists()
        )

    @classmethod
    def _check_upstream_trigger(
        cls, run: Models.Run, op_spec: V1Operation
    ) -> [bool, bool]:
        """
        Checks the upstream and the trigger rule.

        Returns a tuple containing information about if the run:
         * should start
         * can start
        """
        one_done = cls._is_upstream_one_done(run=run)
        # Early opt-out if not upstream is done
        if not one_done:
            return False, False
        if op_spec.trigger == V1TriggerPolicy.ONE_DONE:
            return one_done, one_done
        if op_spec.trigger == V1TriggerPolicy.ONE_SUCCEEDED:
            return one_done, cls._is_upstream_one_succeeded(run=run)
        if op_spec.trigger == V1TriggerPolicy.ONE_FAILED:
            return one_done, cls._is_upstream_one_failed(run=run)

        all_done = cls._is_upstream_all_done(run=run)
        # Early opt-out if not upstream is done
        if not all_done:
            return False, False
        if op_spec.trigger == V1TriggerPolicy.ALL_DONE:
            return all_done, all_done
        # If not trigger policy is set we assume ALL_SUCCEEDED
        if op_spec.trigger is None or op_spec.trigger == V1TriggerPolicy.ALL_SUCCEEDED:
            return all_done, cls._is_upstream_all_succeeded(run=run)
        if op_spec.trigger == V1TriggerPolicy.ALL_FAILED:
            return all_done, cls._is_upstream_all_failed(run=run)

        return False, False

    @staticmethod
    def _trigger_event_downstream(queryset: QuerySet):
        for down_run in queryset:
            workers.send(
                SchedulerCeleryTasks.RUNS_PREPARE,
                kwargs={"run_id": down_run},
                eager_kwargs={"run": down_run},
            )

    @classmethod
    def _trigger_downstream(cls, queryset: QuerySet, is_skipped: bool = False):
        for down_run in queryset:
            op_spec = V1Operation.read(down_run.raw_content)  # TODO: Use construct
            if is_skipped and op_spec.skip_on_upstream_skip:
                condition = V1StatusCondition.get_condition(
                    type=V1Statuses.SKIPPED,
                    status="True",
                    reason="SchedulerNotifyDone",
                    message="Run is skipped because one of the upstream dependencies is skipped.",
                )
                new_run_status(run=down_run, condition=condition)
                continue

            should_start, can_start = cls._check_upstream_trigger(
                run=down_run, op_spec=op_spec
            )
            if should_start:
                if can_start:
                    workers.send(
                        SchedulerCeleryTasks.RUNS_PREPARE,
                        kwargs={"run_id": down_run.id},
                        eager_kwargs={"run": down_run},
                    )
                else:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.UPSTREAM_FAILED,
                        status="True",
                        reason="SchedulerNotifyDone",
                        message="Run could not be started, upstream error.",
                    )
                    new_run_status(run=down_run, condition=condition)

    @staticmethod
    def _should_iterate(run: Models.Run) -> bool:
        if not (
            run.is_matrix
            and (LifeCycle.is_compiled(run.status) or run.status == V1Statuses.RUNNING)
            and run.runtime in V1MatrixKind.iteration_values()
        ):
            return False

        iteration = run.meta_info.get(META_ITERATION)
        if iteration is None:
            return True

        if hasattr(run, "_compiled_operation"):
            compiled_operation = run._compiled_operation
        else:
            compiled_operation = V1CompiledOperation.read(
                run.content
            )  # TODO: Use construct
            # Cache the compiled_operation to avoid recalculation
            run._compiled_operation = compiled_operation
        if compiled_operation.has_hyperband_matrix:
            bracket_iteration = run.meta_info.get(META_BRACKET_ITERATION, 0)
            compiled_operation.matrix.set_tuning_params()
            return compiled_operation.matrix.should_reschedule(
                iteration, bracket_iteration=bracket_iteration
            )
        return compiled_operation.matrix.should_reschedule(iteration)

    @staticmethod
    def _get_max_budget(run: Models.Run) -> int:
        return -1

    @classmethod
    def _start_pipeline(cls, run: Models.Run):
        dag = flows.get_run_dag(run)
        max_budget = 100
        independent_ops = dags.get_independent_ops(dag)
        run_max_budget = cls._get_max_budget(run)
        if len(independent_ops) > 100 and run_max_budget > 0:
            max_budget = run_max_budget
        independent_ops = Models.Run.objects.filter(
            id__in=independent_ops,
            status=V1Statuses.CREATED,
            pending=None,
        ).values_list("id", flat=True)
        for i_run in independent_ops:
            options = {}
            if max_budget < 0:
                options = {"countdown": 2 + abs(max_budget) // 50}
            workers.send(
                SchedulerCeleryTasks.RUNS_PREPARE, kwargs={"run_id": i_run}, **options
            )
            max_budget -= 1

        condition = V1StatusCondition.get_condition(
            type=V1Statuses.RUNNING,
            status="True",
            reason="SchedulerPipeline",
            message="Operation is running",
        )
        new_run_status(run=run, condition=condition)
        return run

    @classmethod
    def _start_run_for_schedule(cls, pipeline: Models.Run, depends_on_past: bool):
        if LifeCycle.is_done(pipeline.status):
            return
        compiled_operation = V1CompiledOperation.read(
            pipeline.content
        )  # TODO: Use construct
        if compiled_operation.schedule.depends_on_past:
            dependence_cond = depends_on_past is True
        else:
            dependence_cond = depends_on_past is False
        if dependence_cond:
            cls.RESOLVER._create_operation_schedule(
                run=pipeline, compiled_operation=compiled_operation
            )

    @classmethod
    def _start_schedule_based_on_run(cls, run: Models.Run, depends_on_past: bool):
        return cls._start_run_for_schedule(
            pipeline=run.pipeline, depends_on_past=depends_on_past
        )

    @classmethod
    def _create_tuner_operation(cls, run):
        run_uuid = run.uuid.hex
        compiled_operation = V1CompiledOperation.read(
            run.content
        )  # TODO: Use construct
        iteration = run.meta_info.get(META_ITERATION)
        params = dict(
            matrix=compiled_operation.matrix,
            iteration=iteration,
        )
        op = None
        if run.runtime == V1MatrixKind.BAYES:
            op = get_bo_tuner(
                tuner=compiled_operation.matrix.tuner,
                join=cls.RESOLVER._get_pipeline_join(
                    pipeline=run_uuid, metric=compiled_operation.matrix.metric
                ),
                **params,
            )
        elif run.runtime == V1MatrixKind.HYPERBAND:
            try:
                bracket_iteration = run.meta_info.get(META_BRACKET_ITERATION)
                additional_filters = ""
                if iteration is not None:
                    additional_filters += "meta_values.iteration:{}".format(iteration)
                if bracket_iteration:
                    additional_filters += ", meta_values.bracket_iteration:{}".format(
                        bracket_iteration
                    )
                op = get_hyperband_tuner(
                    tuner=compiled_operation.matrix.tuner,
                    join=cls.RESOLVER._get_pipeline_join(
                        pipeline=run_uuid,
                        metric=compiled_operation.matrix.metric,
                        additional_filters=additional_filters,
                    ),
                    bracket_iteration=bracket_iteration,
                    **params,
                )
            except ValueError:
                pass
        elif run.runtime == V1MatrixKind.HYPEROPT:
            op = get_hyperopt_tuner(
                tuner=compiled_operation.matrix.tuner,
                join=cls.RESOLVER._get_pipeline_join(
                    pipeline=run_uuid, metric=compiled_operation.matrix.metric
                ),
                **params,
            )
        elif run.runtime == V1MatrixKind.ITERATIVE:
            # TODO
            return

        if op:
            try:
                tuner = operations.init_run(
                    project_id=run.project_id,
                    user_id=run.user_id,
                    pipeline_id=run.id,
                    op_spec=op,
                    controller_id=run.controller_id or run.id,
                    supported_owners={run.project.owner.name},
                ).instance
                tuner.save()
                return tuner
            except (PolyaxonException, ValueError, ValidationError) as e:
                condition = V1StatusCondition.get_condition(
                    type=V1Statuses.FAILED,
                    status="True",
                    reason="SchedulerTunerError",
                    message=f"Could not create a tuner, an exception was raised:\n{e}",
                )
                new_run_status(run=run, condition=condition)
                return
        cls._capture_message(
            "Iteration algorithm could not create a tuning operation: {}".format(run.id)
        )
        condition = V1StatusCondition.get_condition(
            type=V1Statuses.FAILED,
            status="True",
            reason="SchedulerTunerError",
            message="Iteration algorithm could not create a tuning operation.",
        )
        new_run_status(run=run, condition=condition)

    @staticmethod
    def get_run(
        run_id: int,
        run: Optional[Models.Run] = None,
        use_all: bool = False,
        only: Optional[List[str]] = None,
        defer: Optional[List[str]] = None,
        prefetch: Optional[List[str]] = None,
    ) -> Optional[Models.Run]:
        if run:
            return run
        query = Models.Run.all if use_all else Models.Run.objects
        if only:
            query = query.only(*only)
        if defer:
            query = query.only(*defer)
        if prefetch:
            query = query.select_related(*prefetch)

        try:
            return query.get(id=run_id)
        except Models.Run.DoesNotExist:
            _logger.info(
                "Something went wrong, the run `%s` does not exist anymore.", run_id
            )

    @classmethod
    def runs_prepare(
        cls,
        run_id: int,
        run: Optional[Models.Run] = None,
        start: bool = True,
        extra_message: Optional[str] = None,
    ):
        run = cls.get_run(
            run_id=run_id, run=run, prefetch=cls.DEFAULT_PREFETCH + ["controller"]
        )
        if not run:
            return

        if not LifeCycle.is_compilable(run.status):
            _logger.info(
                "Run `%s` cannot transition from `%s` to `%s`.",
                run_id,
                run.status,
                V1Statuses.COMPILED,
            )
            return

        try:
            compiled_at = now()
            _, compiled_operation = cls._resolve(run=run, compiled_at=compiled_at)
        except PolyaxonCompilerError as e:
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.FAILED,
                status="True",
                reason="SchedulerPrepare",
                message=f"Failed to compile.\n{e}",
            )
            new_run_status(run=run, condition=condition)
            return None
        except Exception as e:
            cls._capture_exception(e)
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.FAILED,
                status="True",
                reason="SchedulerPrepare",
                message="Compiler received an unexpected error.",
            )
            new_run_status(run=run, condition=condition)
            return None

        if run.cloning_kind == V1CloningKind.CACHE or run.pending == V1RunPending.BUILD:
            return run

        message = "Run is compiled"
        if extra_message:
            message = f"{message} ({extra_message})."
        condition = V1StatusCondition.get_condition(
            type=V1Statuses.COMPILED,
            status="True",
            reason="SchedulerPrepare",
            message=message,
            last_update_time=compiled_at,
        )
        new_run_status(run=run, condition=condition)

        if cls._should_iterate(run):
            workers.send(
                SchedulerCeleryTasks.RUNS_TUNE,
                kwargs={"run_id": run_id},
                eager_kwargs={"run": run},
            )
            return run
        if start and run.pending is None:
            workers.send(
                SchedulerCeleryTasks.RUNS_START,
                kwargs={"run_id": run_id},
                eager_kwargs={"run": run},
            )
        return run

    @classmethod
    def runs_start_immediately(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(run_id=run_id, run=run)
        if not run:
            return

        if run.managed_by != ManagedBy.AGENT:
            return

        if LifeCycle.is_done(run.status):
            _logger.info(
                "Run `%s` cannot be started from `%s`.",
                run_id,
                run.status,
            )
            return

        if run.is_matrix or run.is_dag:
            return cls._start_pipeline(run)

        condition = V1StatusCondition.get_condition(
            type=V1Statuses.QUEUED,
            status="True",
            reason="SchedulerStart",
            message="Run is queued",
        )
        new_run_status(run=run, condition=condition)
        return run

    @classmethod
    def runs_start(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(run_id=run_id, run=run, prefetch=["pipeline"])
        if not run:
            return

        if not run.is_managed:
            return

        if LifeCycle.is_done(run.status):
            _logger.info(
                "Run `%s` cannot be started from `%s`.",
                run_id,
                run.status,
            )
            return

        if run.schedule_at:
            cls._start_schedule_based_on_run(run, depends_on_past=False)

        if run.is_schedule:
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.RUNNING,
                status="True",
                reason="SchedulerStart",
                message="Schedule is running",
            )
            new_run_status(run=run, condition=condition)
            return None

        if not LifeCycle.is_compiled(run.status):
            _logger.info(
                "Run `%s` cannot transition from `%s` to `%s`.",
                run_id,
                run.status,
                V1Statuses.QUEUED,
            )
            return None

        # Queue the independent or schedule runs immediately
        start_immediately = bool(not run.pipeline_id or run.schedule_at)
        if not start_immediately and run.pipeline_id:
            start_immediately = LifeCycle.is_done(run.pipeline.status)

        if start_immediately:
            cls.runs_start_immediately(run_id=run_id, run=run)
        return run

    @classmethod
    def runs_set_artifacts(
        cls, run_id: int, run: Optional[Models.Run] = None, artifacts: List[Dict] = None
    ):
        if not artifacts:
            return

        run = cls.get_run(run_id=run_id, run=run)
        if not run:
            return

        artifacts = [V1RunArtifact.from_dict(a) for a in artifacts]
        atomic_set_artifacts(run=run, artifacts=artifacts)

    @classmethod
    def runs_stop(
        cls,
        run_id: int,
        run: Optional[Models.Run] = None,
        reason: str = "PipelineStopping",
        message: Optional[str] = None,
    ):
        run = cls.get_run(run_id=run_id, run=run)
        if not run:
            return

        if run.is_managed:
            if run.is_matrix or run.is_dag or run.is_schedule:
                dependent_runs = Models.Run.objects.filter(
                    Q(pipeline_id=run.id) | Q(controller_id=run.id)
                )
                if not dependent_runs.exclude(
                    status__in=LifeCycle.DONE_VALUES
                ).exists():
                    new_run_stop_status(run=run, message=None)
                    return
                # Set the active ops to stopping
                active_pipeline_runs = dependent_runs.filter(
                    status__in=LifeCycle.ON_K8S_VALUES
                )
                if active_pipeline_runs:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.STOPPING,
                        status="True",
                        reason=reason,
                        message="Pipeline controller requested to stop the run.",
                    )
                    bulk_new_run_status(active_pipeline_runs, condition)
                # Set any dependent pipeline with no pipeline
                pipeline_pipeline_runs = get_stopping_pipelines_with_no_runs(
                    dependent_runs
                )
                if pipeline_pipeline_runs:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.STOPPED,
                        status="True",
                        reason=reason,
                        message="Pipeline controller requested to stop the run.",
                    )
                    bulk_new_run_status(pipeline_pipeline_runs, condition)
                # Set the non active / non done ops to stopped
                pipeline_runs = dependent_runs.exclude(
                    status__in=LifeCycle.ON_K8S_VALUES
                    | LifeCycle.DONE_VALUES
                    | {
                        V1Statuses.STOPPING,
                    }
                )
                if pipeline_runs:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.STOPPED,
                        status="True",
                        reason=reason,
                        message="Pipeline controller requested to stop the run.",
                    )
                    bulk_new_run_status(pipeline_runs, condition)
                # If none of the ops is in the stopping mode we mark the operation as stopped
                if not dependent_runs.exclude(
                    status__in=LifeCycle.DONE_VALUES
                ).exists():
                    new_run_stop_status(run=run, message=None)
                    return
            return
        new_run_stop_status(run=run, message=message)

    @classmethod
    def runs_skip(
        cls,
        run_id: int,
        run: Optional[Models.Run] = None,
        reason: str = "PipelineSKipped",
        message: Optional[str] = None,
    ):
        run = cls.get_run(run_id=run_id, run=run)
        if not run:
            return

        if run.is_managed:
            if run.is_matrix or run.is_dag or run.is_schedule:
                dependent_runs = Models.Run.objects.filter(
                    Q(pipeline_id=run.id) | Q(controller_id=run.id)
                )
                if not dependent_runs.exclude(
                    status__in=LifeCycle.DONE_VALUES
                ).exists():
                    new_run_skip_status(run=run, message=None)
                    return
                # Set the active ops to stopping
                active_pipeline_runs = dependent_runs.filter(
                    status__in=LifeCycle.ON_K8S_VALUES
                )
                if active_pipeline_runs:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.STOPPING,
                        status="True",
                        reason=reason,
                        message="Pipeline controller requested to stop the run.",
                    )
                    bulk_new_run_status(active_pipeline_runs, condition)
                # Set any dependent pipeline with no pipeline
                pipeline_pipeline_runs = get_stopping_pipelines_with_no_runs(
                    dependent_runs
                )
                if pipeline_pipeline_runs:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.SKIPPED,
                        status="True",
                        reason=reason,
                        message="Pipeline controller requested to skip the run.",
                    )
                    bulk_new_run_status(pipeline_pipeline_runs, condition)
                # Set the non active / non done ops to stopped
                pipeline_runs = dependent_runs.exclude(
                    status__in=LifeCycle.ON_K8S_VALUES
                    | LifeCycle.DONE_VALUES
                    | {
                        V1Statuses.STOPPING,
                    }
                )
                if pipeline_runs:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.SKIPPED,
                        status="True",
                        reason=reason,
                        message="Pipeline controller requested to skip the run.",
                    )
                    bulk_new_run_status(pipeline_runs, condition)
                # If none of the ops is in the stopping mode we mark the operation as stopped
                if not dependent_runs.exclude(
                    status__in=LifeCycle.DONE_VALUES
                ).exists():
                    new_run_skip_status(run=run, message=None)
                    return
            return
        new_run_skip_status(run=run, message=message)

    @classmethod
    def runs_hooks(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(run_id=run_id, run=run, prefetch=cls.DEFAULT_PREFETCH)
        if not run:
            return

        if not LifeCycle.is_done(run.status):
            _logger.info(
                "Run `%s` has a current status `%s`. "
                "Scheduler cannot execute hooks before the run is done.",
                run_id,
                run.status,
            )
            return None

        try:
            hook_ops = cls._resolve_hooks(run=run)
        except PolyaxonCompilerError as e:
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.FAILED,
                status="True",
                reason="SchedulerHooksResolve",
                message=f"Hooks resolution failed.\n{e}",
            )
            new_run_status(run=run, condition=condition)
            return None
        except Exception as e:
            cls._capture_exception(e)
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.FAILED,
                status="True",
                reason="SchedulerHooksResolve",
                message="Compiler received an unexpected error.",
            )
            new_run_status(run=run, condition=condition)
            return None

        try:
            hook_runs = cls._execute_hooks(run, hook_ops)
        except PolyaxonCompilerError as e:
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.FAILED,
                status="True",
                reason="SchedulerHooksExecute",
                message=f"Hooks execution failed.\n{e}",
            )
            new_run_status(run=run, condition=condition)
            return None
        except Exception as e:
            cls._capture_exception(e)
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.FAILED,
                status="True",
                reason="SchedulerHooksExecute",
                message="Compiler received an unexpected error.",
            )
            new_run_status(run=run, condition=condition)
            return None

        for hook_run_id in hook_runs:
            workers.send(
                SchedulerCeleryTasks.RUNS_PREPARE,
                kwargs={"run_id": hook_run_id},
            )

    @classmethod
    def runs_built(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(run_id=run_id, run=run, prefetch=cls.DEFAULT_PREFETCH)
        if not run:
            return

        if run.status != V1Statuses.CREATED:
            return

        base_approve_run(run)

        if run.pending is not None:
            return

        cls.runs_prepare(
            run_id=run_id,
            run=run,
            extra_message="Image was built successfully",
            start=True,
        )

    @classmethod
    def runs_iterate(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(run_id=run_id, run=run, prefetch=["pipeline"])
        if not run:
            return

        if not run.pipeline_id:
            return

        outputs = run.outputs or {}
        inputs = run.inputs or {}
        suggestions = outputs.get("suggestions", [])
        iteration = inputs.get(META_ITERATION)
        pipeline = run.pipeline

        if not run.outputs or not suggestions or iteration is None:
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.FAILED,
                status="True",
                reason="EmptyIteration",
                message="Run does not have a more iterations and suggestions, early stopped.",
            )
            new_run_status(run=pipeline, condition=condition)
            return

        compiled_operation = cls._set_iteration_meta_info(
            run=pipeline,
            suggestions=suggestions,
            iteration=iteration,
            bracket_iteration=inputs.get(META_BRACKET_ITERATION),
        )
        if compiled_operation:
            runs = cls.RESOLVER._create_iteration_matrix(
                run=pipeline,
                suggestions=suggestions,
                iteration=iteration,
                compiled_operation=compiled_operation,
            )
            if runs:
                # Create edges
                run_edges = [
                    Models.RunEdge(
                        upstream_id=run.id,
                        downstream_id=r,
                        kind=V1RunEdgeKind.JOIN,
                    )
                    for r in runs
                ]
                if run_edges:
                    Models.RunEdge.objects.bulk_create(run_edges)
        return run

    @classmethod
    def runs_tune(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(run_id=run_id, run=run)
        if not run:
            return

        if LifeCycle.is_compiled(run.status):
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.RUNNING,
                status="True",
                reason="SchedulerIterate",
                message="Starting the optimization process.",
            )
            new_run_status(run=run, condition=condition)

        tuner = cls._create_tuner_operation(run)
        if tuner:
            workers.send(
                SchedulerCeleryTasks.RUNS_PREPARE,
                kwargs={"run_id": tuner.id},
                eager_kwargs={"run": tuner},
            )

    @classmethod
    def runs_notify_status(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(
            run_id=run_id,
            run=run,
            only=[
                "id",
                "schedule_at",
                "pipeline_id",
                "controller_id",
                "status",
                "outputs",
                "meta_info",
            ],
        )
        if not run:
            return

        if run.pipeline_id is None:
            return

        # Queue more runs for the pipeline
        downstream_query = run.downstream_edges.filter(
            downstream__status=V1Statuses.CREATED,
            statuses__contains=[run.status],
            pending=None,
        ).values_list("downstream__id", flat=True)
        cls._trigger_event_downstream(downstream_query)

    @classmethod
    def runs_notify_done(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(
            run_id=run_id,
            run=run,
            only=[
                "id",
                "schedule_at",
                "pipeline_id",
                "controller_id",
                "status",
                "outputs",
                "meta_info",
            ],
        )
        if not run:
            return

        cls._notify_cache(run)
        cls._notify_build(run)
        cls._notify_hooks(run)

        if run.pipeline_id is None:
            return

        if run.schedule_at:
            cls._start_schedule_based_on_run(run=run, depends_on_past=True)
            return

        if run.has_tuner_runtime:
            cls.runs_iterate(run_id=run.id, run=run)

        # Queue more runs for the pipeline
        downstream_query = Models.Run.objects.filter(
            upstream_runs=run_id,
            status=V1Statuses.CREATED,
            pending=None,
        ).prefetch_related("upstream_runs")
        cls._trigger_downstream(
            downstream_query, is_skipped=LifeCycle.skipped(run.status)
        )

        if run.downstream_runs.filter(status=V1Statuses.CREATED).count() == 0:
            workers.send(
                SchedulerCeleryTasks.RUNS_CHECK_PIPELINE,
                kwargs={"run_id": run.pipeline_id},
            )
        else:
            workers.send(
                SchedulerCeleryTasks.RUNS_CHECK_EARLY_STOPPING,
                kwargs={"run_id": run.pipeline_id},
            )

    @classmethod
    def runs_wakeup_schedule(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(
            run_id=run_id,
            run=run,
            prefetch=cls.DEFAULT_PREFETCH,
        )
        if not run:
            return

        cls._start_run_for_schedule(pipeline=run, depends_on_past=True)

    @classmethod
    def runs_check_early_stopping(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(run_id=run_id, run=run, defer=STATUS_UPDATE_COLUMNS_DEFER)
        if not run:
            return

        if cls._should_early_stop(run):
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.STOPPING,
                status="True",
                reason="SchedulerEarlyStopping",
                message="Pipeline was early stopped.",
            )
            new_run_status(run=run, condition=condition)
            pipeline_runs = run.pipeline_runs.exclude(status__in=LifeCycle.DONE_VALUES)
            bulk_new_run_status(pipeline_runs, condition)

    @classmethod
    def runs_check_pipeline(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(run_id=run_id, run=run, defer=STATUS_UPDATE_COLUMNS_DEFER)
        if not run:
            return

        def _early_stop():
            message = "Pipeline met an early stopping condition."
            _condition = V1StatusCondition.get_condition(
                type=V1Statuses.STOPPING,
                status="True",
                reason=reason,
                message=message,
            )
            new_run_status(run=run, condition=_condition)
            cls.runs_stop(
                run_id=run_id, run=run, reason="PipelineEarlyStopping", message=message
            )

        reason = "SchedulerCheckPipeline"
        if is_pipeline_done(run):
            if LifeCycle.stopped(run.status):
                return
            elif cls._should_iterate(run):
                if cls._should_early_stop(run):
                    _early_stop()
                else:
                    workers.send(
                        SchedulerCeleryTasks.RUNS_TUNE,
                        kwargs={"run_id": run_id},
                        eager_kwargs={"run": run},
                    )
                return
            elif LifeCycle.is_stopping(run.status):
                condition = V1StatusCondition.get_condition(
                    type=V1Statuses.STOPPED,
                    status="True",
                    reason=reason,
                    message="Pipeline has stopped.",
                )
            else:
                runs_count = get_failed_stopped_and_all_runs(run.id)
                failed_count = runs_count.get("failed", 0)
                stopped_count = runs_count.get("stopped", 0)
                all_count = runs_count.get("all", 0)
                if failed_count > 0:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.FAILED,
                        status="True",
                        reason=reason,
                        message="Pipeline has failed, some operations did not finish successfully.\n"
                        "Number of failed operations: {}".format(runs_count["failed"]),
                    )
                elif stopped_count > 0 and stopped_count == all_count:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.STOPPED,
                        status="True",
                        reason=reason,
                        message="All operations in this pipeline were stopped.",
                    )
                else:
                    condition = V1StatusCondition.get_condition(
                        type=V1Statuses.SUCCEEDED,
                        status="True",
                        reason=reason,
                        message="Pipeline has succeeded.",
                    )
            new_run_status(run=run, condition=condition)
        elif cls._should_early_stop(run):
            _early_stop()

    @classmethod
    def runs_check_orphan_pipeline(cls, run_id: int, run: Optional[Models.Run] = None):
        run = cls.get_run(
            run_id=run_id, run=run, use_all=True, defer=STATUS_UPDATE_COLUMNS_DEFER
        )
        if not run:
            return

        if Models.Run.all.filter(pipeline_id=run_id).count() == 0:
            run.confirm_delete()

    @staticmethod
    def delete_archived_project(project_id: int):
        try:
            project = Models.Project.all.get(id=project_id)
        except Models.Project.DoesNotExist:
            return

        delete_in_progress_project(project)

    @staticmethod
    def delete_archived_run(run_id):
        try:
            run = Models.Run.all.get(id=run_id)
        except Models.Run.DoesNotExist:
            return

        delete_in_progress_run(run)

    @staticmethod
    def stats_calculation_project(project_id: int):
        project = Models.Project.all.select_related("latest_stats").get(id=project_id)
        current_hour = now().replace(minute=0, second=0, microsecond=0)

        run_count = collect_project_run_count_stats(project=project)
        status_count = collect_project_run_status_stats(project=project)
        run_tracking_time = collect_project_run_duration_stats(project=project)
        version_count = collect_project_version_stats(project=project)
        user_count = collect_project_unique_user_stats(project=project)

        is_new = False
        if (
            project.latest_stats is None
            or project.latest_stats.created_at < current_hour
        ):
            is_new = True
            latest_stats = Models.ProjectStats(project=project)
        else:
            latest_stats = project.latest_stats

        latest_stats.run = run_count
        latest_stats.status = status_count
        latest_stats.tracking_time = run_tracking_time
        latest_stats.version = version_count
        latest_stats.user = user_count
        latest_stats.save()

        if is_new:
            project.latest_stats = latest_stats
            project.save(update_fields=["latest_stats"])
        if user_count and user_count.get("ids"):
            user_count["ids"] = list(user_count["ids"])
            project.contributors.add(*user_count["ids"])



from typing import Dict, Optional, Tuple

from haupt.common.settings.apps import set_apps
from haupt.common.settings.celery import set_celery
from haupt.common.settings.core import set_core
from haupt.schemas.platform_config import PlatformConfig


def set_background_service(
    context,
    config: PlatformConfig,
    scheduler_apps: Tuple,
    routes: Dict,
    schedules: Optional[Dict] = None,
    db_app: Optional[str] = None,
):
    set_apps(
        context=context,
        config=config,
        third_party_apps=None,
        project_apps=scheduler_apps,
        db_app=db_app,
    )
    set_core(context=context, config=config, use_db=True)
    set_celery(context=context, config=config, routes=routes, schedules=schedules)


def set_service(context, config: PlatformConfig):
    from haupt.background.celeryp.routes import get_routes

    scheduler_apps = ("haupt.background.scheduler.apps.SchedulerConfig",)
    set_background_service(
        context, config, scheduler_apps=scheduler_apps, routes=get_routes(), db_app=None
    )



from haupt.orchestration.crons.deletion import CronsDeletionManager
from haupt.orchestration.crons.heartbeats import CronsHeartbeatManager
from haupt.orchestration.crons.stats import CronsStatsManager
from haupt.orchestration.scheduler.manager import SchedulingManager


class BackgroundManager(
    SchedulingManager,
    CronsDeletionManager,
    CronsHeartbeatManager,
    CronsStatsManager,
):
    pass



from django.conf import settings

from haupt.common.service_interface import LazyServiceWrapper
from haupt.orchestration.executor.manager import event_manager
from haupt.orchestration.executor.service import ExecutorService


def get_executor_backend_path():
    return settings.EXECUTOR_BACKEND


def get_executor_options():
    return {
        "workers_service": settings.WORKERS_SERVICE,
        "handlers_service": settings.HANDLERS_SERVICE,
    }


backend = LazyServiceWrapper(
    backend_base=ExecutorService,
    backend_path=settings.EXECUTOR_BACKEND,
    options=get_executor_options(),
)
backend.expose(locals())

subscribe = event_manager.subscribe



from typing import Any, Dict, Optional

from haupt.common.options.option import Option


class BaseConfHandler:
    def get(self, option: Option, owners: Optional[Dict[str, int]] = None) -> Any:
        raise NotImplementedError

    def set(self, option: Option, value: Any, owners: Optional[Dict[str, int]] = None):
        raise NotImplementedError

    def delete(self, option: Option, owners: Optional[Dict[str, int]] = None) -> Any:
        raise NotImplementedError



import logging

from typing import List

from polyaxon import settings
from polyaxon._k8s import converter
from polyaxon._k8s.executor.async_executor import AsyncExecutor
from polyaxon._operations import get_notifier_operation
from polyaxon._polyaxonfile import OperationSpecification
from polyaxon.schemas import V1StatusCondition

logger = logging.getLogger("haupt.streams.notification")


async def notify_run(
    namespace: str,
    owner: str,
    project: str,
    run_uuid: str,
    run_name: str,
    condition: V1StatusCondition,
    connections: List[str],
):
    executor = AsyncExecutor(namespace=namespace)
    await executor.refresh()
    for connection in connections:
        connection_type = settings.AGENT_CONFIG.connections_by_names.get(connection)
        if not connection_type:
            logger.warning(
                "Could not create notification using connection {}, "
                "the connection was not found or not set correctly.".format(
                    connection_type
                )
            )
            continue

        operation = get_notifier_operation(
            connection=connection,
            backend=connection_type.kind,
            owner=owner,
            project=project,
            run_uuid=run_uuid,
            run_name=run_name,
            condition=condition.to_dict(),
        )
        compiled_operation = OperationSpecification.compile_operation(operation)
        resource = converter.make(
            owner_name=owner,
            project_name=project,
            project_uuid=project,
            run_uuid=run_uuid,
            run_name=run_name,
            run_path=run_uuid,
            compiled_operation=compiled_operation,
            params=operation.params,
        )
        await executor.create(
            run_uuid=run_uuid,
            run_kind=compiled_operation.get_run_kind(),
            resource=resource,
        )



import logging
import os

from typing import List, Optional

from clipped.utils.lists import to_list

_logger = logging.getLogger("haupt.cli.manage")


def manage(
    command: Optional[str], args: Optional[List[str]] = None, app: Optional[str] = None
):
    from django.core.management import execute_from_command_line

    app = app or "haupt"

    # Required env var to trigger a management command
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", f"{app}.polyconf.settings")

    argv = ["manage.py"]
    if command:
        argv.append(command)
    argv += to_list(args, check_none=True)
    execute_from_command_line(argv)


def migrate(
    migrate_tables: bool = False,
    migrate_db: bool = False,
):
    if migrate_tables:
        _logger.info("Starting tables migration ...")
        manage("tables")
        _logger.info("Tables were migrated correctly!")

    if migrate_db:
        manage("migrate")
        _logger.info("DB Migration finished!")


def run_manage(command: str, args: Optional[List[str]] = None):
    from haupt import settings
    from polyaxon._env_vars.keys import ENV_KEYS_SERVICE, ENV_KEYS_UI_IN_SANDBOX
    from polyaxon._services.values import PolyaxonServices

    if command == "runserver":
        os.environ[ENV_KEYS_SERVICE] = PolyaxonServices.API
    if os.environ.get(ENV_KEYS_SERVICE, PolyaxonServices.API) == PolyaxonServices.API:
        os.environ[ENV_KEYS_UI_IN_SANDBOX] = "true"
        settings.set_sandbox_config()
    manage(command, args)



from typing import List

from django.conf import settings
from django.db.models import Q, QuerySet
from django.utils.timezone import now

from haupt.db.abstracts.runs import BaseRun
from haupt.db.defs import Models
from polyaxon.schemas import LifeCycle, LiveState, V1Statuses


def run_queryset_stopping(queryset):
    queryset.exclude(status__in=LifeCycle.DONE_OR_IN_PROGRESS_VALUES).update(
        status=V1Statuses.STOPPING
    )


def delete_in_progress_project(project: Models.Project):
    if not project.delete_in_progress(update_name=True):
        return False
    Models.Run.all.filter(project=project).exclude(
        live_state=LiveState.DELETION_PROGRESSING
    ).update(live_state=LiveState.DELETION_PROGRESSING)
    # Delete right away
    if settings.HAS_ORG_MANAGEMENT:
        confirm_delete_runs(
            runs=Models.Run.all.filter(project=project, agent__isnull=True)
        )

    return True


def delete_in_progress_run(run: BaseRun):
    if not run.delete_in_progress(set_deleted_at=False):
        return False
    Models.Run.all.filter(Q(pipeline=run) | Q(controller=run)).exclude(
        live_state=LiveState.DELETION_PROGRESSING
    ).update(live_state=LiveState.DELETION_PROGRESSING)
    # Delete right away
    if settings.HAS_ORG_MANAGEMENT:
        confirm_delete_runs(
            runs=Models.Run.all.filter(pipeline=run, agent__isnull=True)
        )
    return True


def archive_project(project: Models.Project):
    if not project.archive():
        return False
    queryset = Models.Run.objects.filter(project=project).exclude(
        live_state=LiveState.ARCHIVED
    )
    run_queryset_stopping(queryset=queryset)
    queryset.update(live_state=LiveState.ARCHIVED, archived_at=now())
    return True


def archive_run(run: BaseRun):
    if not run.archive(commit=False):
        return False
    if not LifeCycle.is_done(run.status, progressing=True):
        run.status = V1Statuses.STOPPING
    run.save(update_fields=["live_state", "status", "archived_at", "updated_at"])
    queryset = Models.Run.objects.filter(
        Q(pipeline_id=run.id) | Q(controller_id=run.id)
    ).exclude(live_state=LiveState.ARCHIVED)
    run_queryset_stopping(queryset=queryset)
    queryset.update(live_state=LiveState.ARCHIVED, archived_at=now())
    return True


def confirm_delete_run(run: BaseRun):
    if not run.confirm_delete():
        return False
    queryset = Models.Run.all.filter(
        Q(pipeline_id=run.id) | Q(controller_id=run.id)
    ).exclude(live_state=LiveState.DELETION_PROGRESSING, deleted_at__isnull=False)
    run_queryset_stopping(queryset=queryset)
    queryset.update(live_state=LiveState.DELETION_PROGRESSING, deleted_at=now())
    return True


def confirm_delete_runs(runs: QuerySet, run_ids: List[int] = None):
    run_ids = run_ids or list(runs.values_list("id", flat=True))
    if not run_ids:
        return
    runs.update(live_state=LiveState.DELETION_PROGRESSING, deleted_at=now())
    Models.Run.all.filter(id__in=run_ids).exclude(
        status__in=LifeCycle.DONE_VALUES
    ).update(status=V1Statuses.STOPPED)
    queryset = Models.Run.all.filter(
        Q(pipeline_id__in=run_ids) | Q(controller_id__in=run_ids)
    ).exclude(live_state=LiveState.DELETION_PROGRESSING, deleted_at__isnull=False)
    run_queryset_stopping(queryset=queryset)
    queryset.update(live_state=LiveState.DELETION_PROGRESSING, deleted_at=now())


def restore_project(project: Models.Project):
    if not project.restore():
        return False
    Models.Run.archived.filter(project=project).update(
        live_state=LiveState.LIVE, archived_at=None, deleted_at=None
    )
    return True


def restore_run(run: BaseRun):
    if not run.restore():
        return False
    Models.Run.archived.filter(pipeline=run).update(
        live_state=LiveState.LIVE, archived_at=None, deleted_at=None
    )
    return True



from haupt.polyconf.settings import PLATFORM_CONFIG

if PLATFORM_CONFIG.is_scheduler_service:
    urlpatterns = []
elif PLATFORM_CONFIG.is_streams_service:
    from haupt.streams.patterns import *  # noqa
else:
    from haupt.apis.patterns import *  # noqa



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0011_alter_artifact_state"),
    ]

    operations = [
        migrations.AlterField(
            model_name="artifact",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="artifactlineage",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="project",
            name="live_state",
            field=models.IntegerField(
                blank=True,
                choices=[
                    (1, "live"),
                    (0, "archived"),
                    (-1, "deletion_progressing"),
                ],
                db_index=True,
                default=1,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="project",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="run",
            name="live_state",
            field=models.IntegerField(
                blank=True,
                choices=[
                    (1, "live"),
                    (0, "archived"),
                    (-1, "deletion_progressing"),
                ],
                db_index=True,
                default=1,
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="run",
            name="updated_at",
            field=models.DateTimeField(auto_now=True, db_index=True),
        ),
    ]



from rest_framework import serializers

from haupt.db.abstracts.runs import BaseRun


class PipelineMixin(serializers.Serializer):
    def get_pipeline(self, obj: BaseRun) -> str:
        if not obj.pipeline_id:
            return None

        return {
            "uuid": obj.pipeline.uuid.hex,
            "name": obj.pipeline.name,
            "kind": obj.pipeline.kind,
        }



from corsheaders.defaults import default_headers

from haupt.schemas.platform_config import PlatformConfig
from polyaxon._services.headers import PolyaxonServiceHeaders


def set_cors(context, config: PlatformConfig):
    # session settings
    context["CORS_ALLOW_CREDENTIALS"] = True
    context["CORS_ALLOWED_ORIGINS"] = config.cors_allowed_origins
    context["CORS_ALLOW_ALL_ORIGINS"] = False if config.cors_allowed_origins else True

    context["CORS_ALLOW_HEADERS"] = (
        default_headers + PolyaxonServiceHeaders.get_headers()
    )

    context["SSL_ENABLED"] = config.ssl_enabled
    context["PROTOCOL"] = "http"
    context["WS_PROTOCOL"] = "ws"
    if config.ssl_enabled:
        context["SESSION_COOKIE_SECURE"] = True
        context["CSRF_COOKIE_SECURE"] = True
        context["SECURE_PROXY_SSL_HEADER"] = ("HTTP_X_FORWARDED_PROTO", "https")
        context["PROTOCOL"] = "https"
        context["WS_PROTOCOL"] = "wss"
    if config.ssl_redirect_enabled:
        context["SECURE_SSL_REDIRECT"] = True



from haupt.common import auditor
from haupt.common.events.registry import bookmark

auditor.subscribe(bookmark.ProjectBookmarkedActorEvent)
auditor.subscribe(bookmark.ProjectUnBookmarkedActorEvent)
auditor.subscribe(bookmark.RunBookmarkedActorEvent)
auditor.subscribe(bookmark.RunUnBookmarkedActorEvent)



from haupt.background.celeryp.tasks import SchedulerCeleryTasks
from haupt.common import conf
from haupt.common.options.registry.core import SCHEDULER_ENABLED
from haupt.db.defs import Models
from haupt.db.managers.cache import get_cache_clones
from haupt.db.managers.live_state import delete_in_progress_run
from haupt.db.managers.runs import add_run_contributors
from haupt.db.managers.statuses import bulk_new_run_status
from haupt.orchestration.scheduler.manager import SchedulingManager
from polyaxon._constants.metadata import (
    META_HAS_DOWNSTREAM_EVENTS_TRIGGER,
    META_UPLOAD_ARTIFACTS,
)
from polyaxon.schemas import (
    LifeCycle,
    ManagedBy,
    V1RunEdgeKind,
    V1RunKind,
    V1RunPending,
    V1StatusCondition,
    V1Statuses,
)


class APIHandler:
    MANAGER = SchedulingManager

    @classmethod
    def handle_run_created(cls, workers_backend, event: "Event") -> None:  # noqa: F821
        """Handles creation, resume, and restart"""
        if event.instance and event.instance.user:
            add_run_contributors(event.instance, users=[event.instance.user])
        # Run is not managed by Polyaxon
        if (
            event.instance and event.instance.managed_by == ManagedBy.USER
        ) or event.data["managed_by"] == ManagedBy.USER:
            return

        # Run is managed by a pipeline
        if (
            event.instance
            and (
                event.instance.pipeline_id is not None
                or event.data.get("pipeline_id") is not None
            )
            and event.instance.status != V1Statuses.RESUMING
        ):
            return

        # Run is pending
        if (
            event.instance
            and event.instance.pending is not None
            and event.instance.pending != V1RunPending.UPLOAD
        ):
            return

        # Eager logic
        eager = False
        if event.instance:
            if (
                event.instance.managed_by == ManagedBy.CLI
                or event.data["managed_by"] == ManagedBy.CLI
            ):
                eager = True
            if (
                not eager
                and event.instance.cloning_kind is None
                and event.instance.status != V1Statuses.RESUMING
                and event.instance.meta_info is not None
                and META_UPLOAD_ARTIFACTS in event.instance.meta_info
            ):
                # Since we are uploading a run, we should automatically resolve the agent if provided
                eager = True

        workers_backend.send(
            SchedulerCeleryTasks.RUNS_PREPARE,
            delay=conf.get(SCHEDULER_ENABLED) and not eager,
            kwargs={"run_id": event.instance_id},
            eager_kwargs={"run": event.instance},
        )

    @classmethod
    def handle_run_approved_triggered(
        cls, workers_backend, event: "Event"
    ) -> None:  # noqa: F821
        run = cls.MANAGER.get_run(run_id=event.instance_id, run=event.instance)
        if not run:
            return

        task = (
            SchedulerCeleryTasks.RUNS_PREPARE
            if run.status == V1Statuses.CREATED
            else SchedulerCeleryTasks.RUNS_START
        )

        workers_backend.send(
            task,
            kwargs={"run_id": event.instance_id},
            eager_kwargs={"run": event.instance},
        )

    @classmethod
    def handle_run_stopped_triggered(
        cls, workers_backend, event: "Event"
    ) -> None:  # noqa: F821
        cls.MANAGER.runs_stop(run_id=event.instance_id, run=event.instance)

    @classmethod
    def handle_run_skipped_triggered(
        cls, workers_backend, event: "Event"
    ) -> None:  # noqa: F821
        cls.MANAGER.runs_skip(run_id=event.instance_id, run=event.instance)

    @classmethod
    def handle_run_done(
        cls, workers_backend, event: "Event" = None
    ) -> None:  # noqa: F821
        """Handles all run done statuses"""
        workers_backend.send(
            SchedulerCeleryTasks.RUNS_NOTIFY_DONE,
            kwargs={"run_id": event.instance_id},
            eager_kwargs={"run": event.instance},
            countdown=2,
        )

    @classmethod
    def handle_run_new_status(
        cls, workers_backend, event: "Event"
    ) -> None:  # noqa: F821
        run = cls.MANAGER.get_run(run_id=event.instance_id, run=event.instance)
        if not run:
            return

        if LifeCycle.is_done(run.status):  # Managed by notify done
            return

        if run.meta_info.get(META_HAS_DOWNSTREAM_EVENTS_TRIGGER):
            workers_backend.send(
                SchedulerCeleryTasks.RUNS_NOTIFY_STATUS,
                kwargs={"run_id": event.instance_id},
                eager_kwargs={"run": event.instance},
            )

    @classmethod
    def handle_new_artifacts(
        cls, workers_backend, event: "Event"
    ) -> None:  # noqa: F821
        artifacts = event.data.get("artifacts")
        if not artifacts:
            return

        workers_backend.send(
            SchedulerCeleryTasks.RUNS_SET_ARTIFACTS,
            kwargs={"run_id": event.instance_id, "artifacts": artifacts},
            eager_kwargs={"run": event.instance},
        )

    @classmethod
    def handle_run_deleted(cls, workers_backend, event: "Event") -> None:  # noqa: F821
        run = cls.MANAGER.get_run(
            run_id=event.instance_id, run=event.instance, use_all=True
        )
        if not run:
            return

        cache_clones = [i for i in get_cache_clones(run.id)]
        if cache_clones:
            Models.Run.all.filter(id__in=cache_clones).update(
                original=None, cloning_kind=None, pending=None
            )
            for cache_clone_id in cache_clones:
                workers_backend.send(
                    SchedulerCeleryTasks.RUNS_PREPARE,
                    kwargs={"run_id": cache_clone_id},
                )

        if run.runtime == V1RunKind.BUILDER:
            awaiting_build = run.downstream_runs.filter(
                upstream_edges__kind=V1RunEdgeKind.BUILD
            )
            condition = V1StatusCondition.get_condition(
                type=V1Statuses.UPSTREAM_FAILED,
                status="True",
                reason="BuildProcessDeleted",
                message=f"Build process {run.uuid.hex} was deleted, "
                f"last build status: {run.status}.",
            )
            bulk_new_run_status(awaiting_build, condition)

        if run.pipeline_id:
            workers_backend.send(
                SchedulerCeleryTasks.RUNS_CHECK_PIPELINE,
                kwargs={"run_id": run.pipeline_id},
            )
        if run.controller_id and run.controller_id != run.pipeline_id:
            workers_backend.send(
                SchedulerCeleryTasks.RUNS_CHECK_PIPELINE,
                kwargs={"run_id": run.controller_id},
            )
        if not run.has_pipeline:
            return

        delete_in_progress_run(run=run)



from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.mixins import CreateModelMixin as DJRCreateModelMixin
from rest_framework.mixins import DestroyModelMixin as DJRDestroyModelMixin
from rest_framework.mixins import ListModelMixin as DRFListModelMixin
from rest_framework.mixins import RetrieveModelMixin as DRFRetrieveModelMixin
from rest_framework.mixins import UpdateModelMixin as DJRUpdateModelMixin
from rest_framework.response import Response


class CreateModelMixin(DJRCreateModelMixin):
    """
    Create a model instance.
    """


class ListModelMixin(DRFListModelMixin):
    """
    List a queryset.
    """

    def list(self, request, *args, **kwargs):
        response = super().list(request, *args, **kwargs)
        self.audit(request, *args, **kwargs)
        return response


class RetrieveModelMixin(DRFRetrieveModelMixin):
    """
    Retrieve a model instance.
    """

    def retrieve(self, request, *args, **kwargs):
        response = super().retrieve(request, *args, **kwargs)
        self.audit(request, *args, **kwargs)
        return response


class UpdateModelMixin(DJRUpdateModelMixin):
    """
    Update a model instance.
    """

    def update(self, request, *args, **kwargs):
        response = super().update(request, *args, **kwargs)
        self.audit(request, *args, **kwargs)
        return response


class DestroyModelMixin(DJRDestroyModelMixin):
    """
    Destroy a model instance.
    """

    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        self.audit(request, *args, **kwargs)
        self.perform_destroy(instance)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def perform_destroy(self, instance):
        return super().perform_destroy(instance)


class StatsMixin:
    def validate_stats_mode(self):
        mode = self.request.query_params.get("mode")
        if mode not in {"stats", "analytics"}:
            raise ValidationError("Received an unsupported mode: {}".format(mode))
        return mode



from typing import Dict, List, Optional, Tuple

from django.db.models import QuerySet

from haupt.db.abstracts.runs import BaseRun
from haupt.db.defs import Models
from polyaxon.schemas import dags


def get_run_upstream(
    run: BaseRun, id_field: str = "id", pipeline_id: Optional[int] = None
) -> QuerySet:
    upstream_runs = run.upstream_runs
    if pipeline_id:
        upstream_runs = upstream_runs.filter(pipeline_id=pipeline_id)
    return upstream_runs.values_list(id_field, flat=True)


def get_run_downstream(
    run: BaseRun, id_field: str = "id", pipeline_id: Optional[int] = None
) -> QuerySet:
    downstream_runs = run.downstream_runs
    if pipeline_id:
        downstream_runs = downstream_runs.filter(pipeline_id=pipeline_id)
    return downstream_runs.values_list(id_field, flat=True)


def get_run_dag(run: BaseRun) -> Tuple[Dict, Dict]:
    runs = run.pipeline_runs.only("id", "pipeline_id", "uuid").prefetch_related(
        "upstream_runs"
    )
    ops = ((run, get_run_upstream(run)) for run in runs)
    return dags.process_dag(ops)


def get_run_graph(filters) -> Dict[str, List[str]]:
    runs = (
        Models.Run.objects.filter(**filters)
        .only("id", "pipeline_id", "uuid")
        .prefetch_related("downstream_runs")
    )
    return {
        run.uuid.hex: {
            "downstream": [uid.hex for uid in get_run_downstream(run, "uuid")]
        }
        for run in runs
    }



from rest_framework.exceptions import ValidationError
from rest_framework.response import Response

from django.conf import settings
from django.http import Http404

from haupt.apis.bookmarks.views import BookmarkCreateView, BookmarkDeleteView
from haupt.apis.endpoints.run import RunEndpoint
from haupt.apis.methods import runs as methods
from haupt.apis.serializers.runs import (
    RunDetailSerializer,
    RunSerializer,
    RunStatusSerializer,
)
from haupt.common import conf
from haupt.common.apis.regex import RUN_UUID_KEY
from haupt.common.content_types import ContentTypes
from haupt.common.endpoints.base import (
    CreateEndpoint,
    DestroyEndpoint,
    PostEndpoint,
    RetrieveEndpoint,
    UpdateEndpoint,
)
from haupt.common.endpoints.mixins import StatsMixin
from haupt.common.events.registry.archive import RUN_ARCHIVED_ACTOR, RUN_RESTORED_ACTOR
from haupt.common.events.registry.bookmark import (
    RUN_BOOKMARKED_ACTOR,
    RUN_UNBOOKMARKED_ACTOR,
)
from haupt.common.events.registry.run import (
    RUN_APPROVED_ACTOR,
    RUN_COPIED_ACTOR,
    RUN_DELETED_ACTOR,
    RUN_INVALIDATED_ACTOR,
    RUN_RESTARTED_ACTOR,
    RUN_RESUMED_ACTOR,
    RUN_SKIPPED_ACTOR,
    RUN_STATS_ACTOR,
    RUN_STOPPED_ACTOR,
    RUN_TRANSFERRED_ACTOR,
)
from haupt.common.options.registry.k8s import K8S_NAMESPACE
from haupt.db.defs import Models
from haupt.db.managers.live_state import delete_in_progress_run
from haupt.db.managers.stats import StatsSerializer
from haupt.db.queries.runs import STATUS_UPDATE_COLUMNS_DEFER
from haupt.orchestration import operations
from polyaxon._services.values import PolyaxonServices
from polyaxon.schemas import LifeCycle, V1CloningKind


class RunDetailView(RunEndpoint, RetrieveEndpoint, DestroyEndpoint, UpdateEndpoint):
    serializer_class = RunDetailSerializer
    AUDITOR_EVENT_TYPES = {
        "DELETE": RUN_DELETED_ACTOR,
    }
    ALLOWED_METHODS = ["GET", "PUT", "PATCH", "DELETE"]
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True
    AUDIT_OWNER = True
    AUDIT_PROJECT = True

    def _get_related_fields(self):
        return ["original", "pipeline", "project"]

    def _get_prefetch_fields(self):
        return []

    def get_queryset(self):
        if self.request.method == "GET":
            prefetch_fields = self._get_prefetch_fields()
            queryset = Models.Run.all.select_related(*self._get_related_fields())
            if prefetch_fields:
                queryset = queryset.prefetch_related(*prefetch_fields)
            return queryset
        return super().get_queryset()

    def perform_destroy(self, instance):
        delete_in_progress_run(instance)


class RunCloneView(RunEndpoint, CreateEndpoint):
    serializer_class = RunSerializer
    ALLOWED_METHODS = ["POST"]
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True
    AUDIT_PROJECT = True
    AUDIT_OWNER = True

    def clone(self, obj, content, **kwargs):
        pass

    def pre_validate(self, obj):
        # Restart/Copy original run
        if obj.cloning_kind == V1CloningKind.CACHE:
            return obj.original
        return obj

    def post(self, request, *args, **kwargs):
        return methods.clone_run(view=self, request=request, *args, **kwargs)


class RunRestartView(RunCloneView):
    AUDITOR_EVENT_TYPES = {"POST": RUN_RESTARTED_ACTOR}

    def _get_Additional_fields(self, obj):
        return {}

    def clone(self, obj, content, **kwargs):
        return operations.restart_run(
            run=obj,
            user_id=self.request.user.id,
            content=content,
            name=kwargs.get("name"),
            description=kwargs.get("description"),
            tags=kwargs.get("tags"),
            meta_info=kwargs.get("meta_info"),
            **self._get_Additional_fields(obj)
        )


class RunResumeView(RunCloneView):
    AUDITOR_EVENT_TYPES = {"POST": RUN_RESUMED_ACTOR}

    def _get_additional_fields(self, obj):
        return {}

    def clone(self, obj, content, **kwargs):
        return operations.resume_run(
            run=obj,
            user_id=self.request.user.id,
            content=content,
            name=kwargs.get("name"),
            description=kwargs.get("description"),
            tags=kwargs.get("tags"),
            meta_info=kwargs.get("meta_info"),
            message="Run was resumed by user.",
            **self._get_additional_fields(obj)
        )

    def pre_validate(self, obj):
        if not LifeCycle.is_done(obj.status):
            raise ValidationError(
                "Cannot resume this run, the run must reach a final state first, "
                "current status error: {}".format(obj.status)
            )

        if obj.cloning_kind == V1CloningKind.CACHE:
            raise ValidationError(
                "Cannot resume this run, this run is a cache hit and was never executed."
            )
        return super().pre_validate(obj)


class RunCopyView(RunCloneView):
    AUDITOR_EVENT_TYPES = {"POST": RUN_COPIED_ACTOR}

    def clone(self, obj, content, **kwargs):
        return operations.copy_run(
            run=obj,
            user_id=self.request.user.id,
            content=content,
            name=kwargs.get("name"),
            description=kwargs.get("description"),
            tags=kwargs.get("tags"),
            meta_info=kwargs.get("meta_info"),
        )


class RunStatusListView(RunEndpoint, RetrieveEndpoint, CreateEndpoint):
    queryset = Models.Run.restorable.defer(*STATUS_UPDATE_COLUMNS_DEFER).select_related(
        "project",
    )
    serializer_class = RunStatusSerializer
    throttle_scope = "run-status"
    ALLOWED_METHODS = ["POST", "GET"]

    def perform_create(self, serializer):
        try:
            methods.create_status(view=self, serializer=serializer)
        except Models.Run.DoesNotExist:
            raise Http404


class RunStopView(RunEndpoint, CreateEndpoint):
    ALLOWED_METHODS = ["POST"]
    AUDITOR_EVENT_TYPES = {"POST": RUN_STOPPED_ACTOR}
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True
    AUDIT_PROJECT_RESOURCES = True
    AUDIT_PROJECT = True
    AUDIT_OWNER = True

    def post(self, request, *args, **kwargs):
        return methods.stop_run(view=self, request=request, *args, **kwargs)


class RunSkipView(RunEndpoint, CreateEndpoint):
    ALLOWED_METHODS = ["POST"]
    AUDITOR_EVENT_TYPES = {"POST": RUN_SKIPPED_ACTOR}
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True
    AUDIT_PROJECT_RESOURCES = True
    AUDIT_PROJECT = True
    AUDIT_OWNER = True

    def post(self, request, *args, **kwargs):
        return methods.skip_run(view=self, request=request, *args, **kwargs)


class RunApproveView(RunEndpoint, CreateEndpoint):
    ALLOWED_METHODS = ["POST"]
    AUDITOR_EVENT_TYPES = {"POST": RUN_APPROVED_ACTOR}
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True

    def post(self, request, *args, **kwargs):
        return methods.approve_run(view=self, request=request, *args, **kwargs)


class RunNamespaceView(RunEndpoint, RetrieveEndpoint):
    ALLOWED_METHODS = ["GET"]

    def retrieve(self, request, *args, **kwargs):
        namespace = {"namespace": conf.get(K8S_NAMESPACE)}
        return Response(namespace)


class RunTransferView(RunEndpoint, CreateEndpoint):
    ALLOWED_METHODS = ["POST"]
    AUDITOR_EVENT_TYPES = {"POST": RUN_TRANSFERRED_ACTOR}
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True

    def post(self, request, *args, **kwargs):
        return methods.transfer_run(view=self, request=request, *args, **kwargs)


class RunInvalidateView(RunEndpoint, PostEndpoint):
    serializer_class = RunSerializer
    AUDITOR_EVENT_TYPES = {"POST": RUN_INVALIDATED_ACTOR}
    ALLOWED_METHODS = ["POST"]
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True

    def post(self, request, *args, **kwargs):
        return methods.invalidate_run(view=self, request=request, *args, **kwargs)


class RunArchiveView(RunEndpoint, PostEndpoint):
    AUDITOR_EVENT_TYPES = {"POST": RUN_ARCHIVED_ACTOR}
    ALLOWED_METHODS = ["POST"]
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True

    def post(self, request, *args, **kwargs):
        return methods.archive_run(view=self, request=request, *args, **kwargs)


class RunRestoreView(RunEndpoint, PostEndpoint):
    AUDITOR_EVENT_TYPES = {"POST": RUN_RESTORED_ACTOR}
    ALLOWED_METHODS = ["POST"]
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True

    def post(self, request, *args, **kwargs):
        return methods.restore_run(view=self, request=request, *args, **kwargs)


SELECT_RELATED = ["project__owner"] if settings.HAS_ORG_MANAGEMENT else []


class RunBookmarkCreateView(RunEndpoint, BookmarkCreateView):
    queryset = Models.Run.restorable.prefetch_related(*SELECT_RELATED, "project")
    AUDITOR_EVENT_TYPES = {"POST": RUN_BOOKMARKED_ACTOR}
    ALLOWED_METHODS = ["POST"]
    content_type = ContentTypes.RUN.value
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True


class RunBookmarkDeleteView(RunEndpoint, BookmarkDeleteView):
    queryset = Models.Run.restorable.prefetch_related(*SELECT_RELATED, "project")
    AUDITOR_EVENT_TYPES = {"DELETE": RUN_UNBOOKMARKED_ACTOR}
    ALLOWED_METHODS = ["DELETE"]
    content_type = ContentTypes.RUN.value
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True


class RunStatsView(RunEndpoint, RetrieveEndpoint, StatsMixin):
    AUDITOR_EVENT_TYPES = {
        "GET": RUN_STATS_ACTOR,
    }
    ALLOWED_METHODS = ["GET"]
    ALLOWED_SERVICES = [PolyaxonServices.UI]
    CHECK_SERVICE = True
    AUDIT_PROJECT = True
    AUDIT_OWNER = True
    AUDIT_PROJECT_RESOURCES = True
    PROJECT_RESOURCE_KEY = RUN_UUID_KEY
    AUDIT_INSTANCE = True
    bookmarked_model = "run"

    def get_serializer(self, *args, **kwargs):
        self.validate_stats_mode()
        queryset = Models.Run.restorable.filter(pipeline_id=self.run.id)
        queryset = StatsSerializer.filter_queryset(
            queryset=queryset,
            request=self.request,
            view=self,
        )
        return StatsSerializer(
            queryset=queryset,
            kind=self.request.query_params.get("kind"),
            aggregate=self.request.query_params.get("aggregate"),
            groupby=self.request.query_params.get("groupby"),
            trunc=self.request.query_params.get("trunc"),
        )



from rest_framework.generics import get_object_or_404

from django.conf import settings
from django.http import HttpRequest

from haupt.common.apis.regex import (
    NAME_KEY,
    OWNER_NAME_KEY,
    PROJECT_NAME_KEY,
    PROJECT_OWNER_NAME_KEY,
    UUID_KEY,
    VERSION_NAME_KEY,
)
from haupt.common.endpoints.base import BaseEndpoint
from haupt.common.permissions import PERMISSIONS_MAPPING
from haupt.db.defs import Models
from polyaxon.schemas import LiveState


class ProjectEndpoint(BaseEndpoint):
    queryset = (
        Models.Project.all.select_related("owner")
        if settings.HAS_ORG_MANAGEMENT
        else Models.Project.all
    )
    lookup_field = NAME_KEY
    lookup_url_kwarg = PROJECT_NAME_KEY
    permission_classes = PERMISSIONS_MAPPING.get(["PROJECT_PERMISSION"])
    CONTEXT_KEYS = (OWNER_NAME_KEY, PROJECT_NAME_KEY)
    CONTEXT_OBJECTS = ("project",)

    PROJECT_NAME_KEY = "name"
    PROJECT_OWNER_NAME_KEY = OWNER_NAME_KEY

    def enrich_queryset(self, queryset):
        if settings.HAS_ORG_MANAGEMENT:
            return queryset.filter(owner__name=self.owner_name)
        return queryset

    def initialize_object_context(self, request: HttpRequest, *args, **kwargs):
        #  pylint:disable=attribute-defined-outside-init
        self.project = self.get_object()

    def set_owner(self):
        self._owner_id = self.project.owner.id


class ProjectResourceListEndpoint(ProjectEndpoint):
    permission_classes = PERMISSIONS_MAPPING.get(["PROJECT_RESOURCE_PERMISSION"])
    AUDITOR_EVENT_TYPES = None

    def _get_object(self):
        filters = {"name": self.project_name}
        if settings.HAS_ORG_MANAGEMENT:
            filters["owner__name"] = self.owner_name
            filters["owner__live_state"] = LiveState.LIVE
        return get_object_or_404(
            Models.Project,
            **filters,
        )

    def get_object(self):
        if self._object:
            return self._object
        self._object = self._get_object()
        if self.permission_classes:
            permission = self.permission_classes[0]()
            cond = permission.has_object_permission(
                request=self.request, view=self, obj=self._object
            )
            if not cond:
                self.permission_denied(
                    self.request, message=getattr(permission, "message", None)
                )
        return self._object

    def enrich_queryset(self, queryset):
        return queryset.filter(project=self.project)


class ProjectResourceEndpoint(ProjectEndpoint):
    permission_classes = PERMISSIONS_MAPPING.get(["PROJECT_RESOURCE_PERMISSION"])
    lookup_field = UUID_KEY
    AUDITOR_EVENT_TYPES = None

    PROJECT_NAME_KEY = PROJECT_NAME_KEY
    PROJECT_OWNER_NAME_KEY = PROJECT_OWNER_NAME_KEY


class VersionListEndpoint(ProjectResourceListEndpoint):
    pass


class VersionEndpoint(ProjectResourceEndpoint):
    permission_classes = PERMISSIONS_MAPPING.get(["PROJECT_VERSION_PERMISSION"])
    lookup_field = NAME_KEY
    lookup_url_kwarg = VERSION_NAME_KEY
    CONTEXT_KEYS = (OWNER_NAME_KEY, PROJECT_NAME_KEY, VERSION_NAME_KEY)
    CONTEXT_OBJECTS = ("version",)

    def enrich_queryset(self, queryset):
        org_filters = (
            {"project__owner__name": self.owner_name}
            if settings.HAS_ORG_MANAGEMENT
            else {}
        )
        return queryset.filter(project__name=self.project_name, **org_filters)

    def set_owner(self):
        self._owner_id = self.version.project.owner_id

    def initialize_object_context(self, request: HttpRequest, *args, **kwargs) -> None:
        #  pylint:disable=attribute-defined-outside-init
        self.version = self.get_object()
        self.set_owner()
